from itertools import islice
from pathlib import Path, PurePath, PurePosixPath
from typing import Any, List, Optional, Set, Tuple, Union

import anndata as ad
import lamindb_setup
import pandas as pd
from anndata import AnnData
from appdirs import AppDirs
from lamin_utils import colors, logger
from lamindb_setup import settings as setup_settings
from lamindb_setup._init_instance import register_storage
from lamindb_setup.dev import StorageSettings
from lamindb_setup.dev._docs import doc_args
from lnschema_core import Feature, FeatureSet, File, Run, Storage, ids
from lnschema_core.types import AnnDataLike, DataLike, FieldAttr, PathLike

from lamindb.dev import run_context
from lamindb.dev._settings import settings
from lamindb.dev.hashing import b16_to_b64, hash_file
from lamindb.dev.storage import (
    UPath,
    delete_storage,
    infer_suffix,
    load_to_memory,
    size_adata,
    write_to_file,
)
from lamindb.dev.storage._backed_access import AnnDataAccessor, BackedAccessor
from lamindb.dev.storage.file import (
    ProgressCallback,
    _str_to_path,
    auto_storage_key_from_file,
    auto_storage_key_from_id_suffix,
    extract_suffix_from_path,
    filepath_from_file,
)
from lamindb.dev.utils import attach_func_to_class_method

from . import _TESTING
from ._feature import convert_numpy_dtype_to_lamin_feature_type
from ._parents import view_lineage
from .dev.storage.file import AUTO_KEY_PREFIX

DIRS = AppDirs("lamindb", "laminlabs")


def process_pathlike(
    filepath: Union[Path, UPath], skip_existence_check: bool = False
) -> Tuple[Storage, bool]:
    if not skip_existence_check:
        try:  # check if file exists
            if not filepath.exists():
                raise FileNotFoundError(filepath)
        except PermissionError:
            pass
    if not isinstance(filepath, UPath):
        filepath = filepath.resolve()
    # check whether the path is in default storage
    default_storage = lamindb_setup.settings.storage.record
    if check_path_is_child_of_root(filepath, default_storage.root_as_path()):
        use_existing_storage_key = True
        return default_storage, use_existing_storage_key
    else:
        # check whether the path is part of one of the existing
        # already-registered storage locations
        result = check_path_in_existing_storage(filepath)
        if isinstance(result, Storage):
            use_existing_storage_key = True
            return result, use_existing_storage_key
        else:
            # if the path is in the cloud, we have a good candidate
            # for the storage root: the bucket
            if isinstance(filepath, UPath):
                # for a UPath, new_root is always the bucket name
                new_root = list(filepath.parents)[-1]
                new_root_str = new_root.as_posix()
                logger.warning(
                    f"creating new storage location for root: {new_root_str}"
                )
                storage_settings = StorageSettings(new_root_str)
                register_storage(storage_settings)
                use_existing_storage_key = True
                return storage_settings.record, use_existing_storage_key
            # if the filepath is local
            else:
                use_existing_storage_key = False
                # if the default storage is local we'll throw an error if the user
                # doesn't provide a key
                if not lamindb_setup.settings.storage.is_cloud:
                    return default_storage, use_existing_storage_key
                # if the default storage is in the cloud (the file is going to
                # be uploaded upon saving it), we treat the filepath as a cache
                else:
                    return default_storage, use_existing_storage_key


def process_data(
    provisional_id: str,
    data: Union[PathLike, DataLike],
    format,
    skip_existence_check: bool = False,
) -> Tuple[Any, Union[Path, UPath], str, Storage, bool]:
    """Serialize a data object that's provided as file or in memory."""
    # if not overwritten, data gets stored in default storage
    if isinstance(data, (str, Path, UPath)):  # PathLike, spelled out
        filepath = (
            data if isinstance(data, (Path, UPath)) else _str_to_path(data)
        )  # returns Path for local
        storage, use_existing_storage_key = process_pathlike(
            filepath, skip_existence_check=skip_existence_check
        )
        suffix = extract_suffix_from_path(filepath)
        memory_rep = None
    elif isinstance(data, (pd.DataFrame, AnnData)):  # DataLike, spelled out
        storage = lamindb_setup.settings.storage.record
        memory_rep = data
        suffix = infer_suffix(data, format)
        cache_name = f"{provisional_id}{suffix}"
        if lamindb_setup.settings.storage.cache_dir is not None:
            filepath = lamindb_setup.settings.storage.cache_dir / cache_name
        else:
            # this should likely be added to lamindb_setup.settings.storage
            cache_dir = Path(DIRS.user_cache_dir)
            cache_dir.mkdir(parents=True, exist_ok=True)
            filepath = cache_dir / cache_name
        # Alex: I don't understand the line below
        if filepath.suffixes == []:
            filepath = filepath.with_suffix(suffix)
        if suffix != ".zarr":
            write_to_file(data, filepath)
        use_existing_storage_key = False
    else:
        raise NotImplementedError(
            f"Do not know how to create a file object from {data}, pass a filepath"
            " instead!"
        )
    return memory_rep, filepath, suffix, storage, use_existing_storage_key


def get_hash(
    filepath,
    suffix,
    filepath_stat=None,
    check_hash: bool = True,
) -> Union[Tuple[Optional[str], Optional[str]], File]:
    if suffix in {".zarr", ".zrad"}:
        return None
    if isinstance(filepath, UPath):
        stat = filepath_stat
        if stat is not None and "ETag" in stat:
            # small files
            if "-" not in stat["ETag"]:
                # only store hash for non-multipart uploads
                # we can't rapidly validate multi-part uploaded files client-side
                # we can add more logic later down-the-road
                hash = b16_to_b64(stat["ETag"])
                hash_type = "md5"
            else:
                stripped_etag, suffix = stat["ETag"].split("-")
                suffix = suffix.strip('"')
                hash = f"{b16_to_b64(stripped_etag)}-{suffix}"
                hash_type = "md5-n"  # this is the S3 chunk-hashing strategy
        else:
            logger.warning(f"did not add hash for {filepath}")
            return None, None
    else:
        hash, hash_type = hash_file(filepath)
    if not check_hash:
        return hash, hash_type
    result = File.filter(hash=hash).list()
    if len(result) > 0:
        if settings.upon_file_create_if_hash_exists == "error":
            msg = f"file with same hash exists: {result[0]}"
            hint = (
                "💡 you can make this error a warning:\n"
                "    ln.settings.upon_file_create_if_hash_exists"
            )
            raise RuntimeError(f"{msg}\n{hint}")
        elif settings.upon_file_create_if_hash_exists == "warn_create_new":
            logger.warning(
                "creating new File object despite existing file with same hash:"
                f" {result[0]}"
            )
            return hash, hash_type
        else:
            logger.warning(f"returning existing file with same hash: {result[0]}")
            return result[0]
    else:
        return hash, hash_type


def get_run(run: Optional[Run]) -> Optional[Run]:
    if run is None:
        run = run_context.run
        if run is None:
            logger.hint(
                "no run & transform get linked, consider passing a `run` or calling"
                " ln.track()"
            )
    return run


def get_path_size_hash(
    filepath: Union[Path, UPath],
    memory_rep: Optional[Union[pd.DataFrame, AnnData]],
    suffix: str,
    check_hash: bool = True,
):
    cloudpath = None
    localpath = None
    hash_and_type: Tuple[Optional[str], Optional[str]]

    if suffix == ".zarr":
        if memory_rep is not None:
            size = size_adata(memory_rep)
        else:
            if isinstance(filepath, UPath):
                cloudpath = filepath
                # todo: properly calculate size
                size = 0
            else:
                localpath = filepath
                size = sum(
                    f.stat().st_size for f in filepath.rglob("*") if f.is_file()  # type: ignore # noqa
                )
        hash_and_type = None, None
    else:
        # to accelerate ingesting high numbers of files
        if settings.upon_file_create_skip_size_hash:
            size = None
            hash_and_type = None, None
        else:
            filepath_stat = filepath.stat()
            if isinstance(filepath, UPath):
                try:
                    size = filepath_stat["size"]  # type: ignore
                # here trying to fix access issue with new s3 buckets
                except Exception as e:
                    if filepath._url.scheme == "s3":
                        filepath = UPath(filepath, cache_regions=True)
                        size = filepath_stat["size"]  # type: ignore
                    else:
                        raise e
                cloudpath = filepath
                hash_and_type = None, None
            else:
                size = filepath_stat.st_size  # type: ignore
                localpath = filepath
            hash_and_type = get_hash(
                filepath, suffix, filepath_stat=filepath_stat, check_hash=check_hash
            )
    return localpath, cloudpath, size, hash_and_type


def check_path_in_existing_storage(
    filepath: Union[Path, UPath]
) -> Union[Storage, bool]:
    for storage in Storage.filter().all():
        # if path is part of storage, return it
        if check_path_is_child_of_root(filepath, root=_str_to_path(storage.root)):
            return storage
    return False


def check_path_is_child_of_root(
    filepath: Union[Path, UPath], root: Optional[Union[Path, UPath]] = None
) -> bool:
    assert isinstance(filepath, Path)
    if root is None:
        root = lamindb_setup.settings.storage.root
    # the following comparisons can fail if types aren't comparable
    if isinstance(filepath, UPath) and isinstance(root, UPath):
        # the following tests equivalency of two UPath objects
        # via string representations; otherwise
        # S3Path('s3://lndb-storage/') and S3Path('s3://lamindb-ci/')
        # test as equivalent
        return list(filepath.parents)[-1].as_posix() == root.as_posix()
    elif not isinstance(filepath, UPath) and not isinstance(root, UPath):
        return root.resolve() in filepath.resolve().parents
    else:
        return False


def get_relative_path_to_directory(
    path: Union[PurePath, Path, UPath], directory: Union[PurePath, Path, UPath]
) -> Union[PurePath, Path]:
    if isinstance(directory, UPath):
        # UPath.relative_to() is not behaving as it should (2023-04-07)
        relpath = PurePath(path.as_posix().replace(directory.as_posix(), ""))
    elif isinstance(directory, Path):
        relpath = path.resolve().relative_to(directory.resolve())  # type: ignore
    elif isinstance(directory, PurePath):
        relpath = path.relative_to(directory)
    else:
        raise TypeError("Directory not of type Path or UPath")
    return relpath


def get_file_kwargs_from_data(
    *,
    data: Union[Path, UPath, str, pd.DataFrame, AnnData],
    key: Optional[str],
    run: Optional[Run],
    format: Optional[str],
    provisional_id: str,
    skip_check_exists: bool = False,
):
    run = get_run(run)
    memory_rep, filepath, suffix, storage, use_existing_storage_key = process_data(
        provisional_id, data, format, skip_check_exists
    )
    # the following will return a localpath that is not None if filepath is local
    # it will return a cloudpath that is not None if filepath is on the cloud
    local_filepath, cloud_filepath, size, hash_and_type = get_path_size_hash(
        filepath,
        memory_rep,
        suffix,
    )
    if isinstance(hash_and_type, File):
        return hash_and_type, None
    else:
        hash, hash_type = hash_and_type

    check_path_in_storage = False
    if key is None and use_existing_storage_key:
        key = get_relative_path_to_directory(
            path=filepath, directory=storage.root_as_path()
        ).as_posix()
        check_path_in_storage = True
    else:
        storage = lamindb_setup.settings.storage.record

    if key is not None and key.startswith(AUTO_KEY_PREFIX):
        raise ValueError(f"Key cannot start with {AUTO_KEY_PREFIX}")

    log_storage_hint(
        check_path_in_storage=check_path_in_storage,
        storage=storage,
        key=key,
        id=provisional_id,
        suffix=suffix,
    )

    kwargs = dict(
        suffix=suffix,
        hash=hash,
        hash_type=hash_type,
        key=key,
        size=size,
        storage_id=storage.id,
        # passing both the id and the object
        # to make them both available immediately
        # after object creation
        run_id=run.id if run is not None else None,
        run=run,
    )
    privates = dict(
        local_filepath=local_filepath,
        cloud_filepath=cloud_filepath,
        memory_rep=memory_rep,
        check_path_in_storage=check_path_in_storage,
    )

    return kwargs, privates


def log_storage_hint(
    *,
    check_path_in_storage: bool,
    storage: Optional[Storage],
    key: Optional[str],
    id: str,
    suffix: str,
) -> None:
    hint = ""
    if check_path_in_storage:
        display_root = storage.root  # type: ignore
        # check whether path is local
        if not storage.root.startswith(("s3://", "gs://")):  # type: ignore
            # if it's a local path, check whether it's in the current working directory
            root_path = Path(storage.root)  # type: ignore
            if check_path_is_child_of_root(root_path, Path.cwd()):
                # only display the relative path, not the fully resolved path
                display_root = root_path.relative_to(Path.cwd())
        hint += f"file in storage '{display_root}'"  # type: ignore
    else:
        hint += "file will be copied to default storage upon `save()`"
    if key is None:
        storage_key = auto_storage_key_from_id_suffix(id, suffix)
        hint += f" with key `None` ('{storage_key}')"
    else:
        hint += f" with key '{key}'"
    logger.hint(hint)


def data_is_anndata(data: DataLike):
    if isinstance(data, AnnData):
        return True
    if isinstance(data, (str, Path, UPath)):
        return Path(data).suffix in {".h5ad", ".zrad"}
    return False  # pragma: no cover


def data_is_mudata(data: DataLike):  # pragma: no cover
    try:
        from mudata import MuData
    except ModuleNotFoundError:
        return False

    if isinstance(data, MuData):
        return True
    if isinstance(data, (str, Path, UPath)):
        return Path(data).suffix in {".h5mu"}
    return False


# uses `initial_version_id` to extract a stem_id that's part of id
# this entire piece of logic might be removed in the future if it doesn't turn out
# to be robustly maintainable
def init_id(
    *,
    provisional_id: Optional[None] = None,
    initial_version_id: Optional[str] = None,
    version: Optional[str] = None,
) -> str:
    if version is not None:
        if not isinstance(version, str):
            raise ValueError(
                "`version` parameter must be `None` or `str`, e.g., '0', '1', etc."
            )
    if initial_version_id is not None:
        stem_id = initial_version_id[:18]
    else:
        stem_id = None
    # first consider an unversioned record
    if version is None and stem_id is None:
        provisional_id = ids.base62_20()
        return provisional_id  # type: ignore
    # now consider a versioned record
    id_ext = ids.base62(2)
    if provisional_id is None and stem_id is None:
        stem_id = ids.base62_18()
        provisional_id = stem_id + id_ext
    elif stem_id is not None:
        assert isinstance(stem_id, str) and len(stem_id) == 18
        provisional_id = stem_id + id_ext
    elif provisional_id is not None:
        assert isinstance(provisional_id, str) and len(provisional_id) == 20
        stem_id = provisional_id[:18]
    return provisional_id  # type: ignore


def set_version(version: Optional[str] = None, previous_version: Optional[str] = None):
    """(Auto-) set version.

    If `version` is `None`, returns the stored version.
    Otherwise sets the version to the passed version.

    Args:
        version: Version string.
        stored_version: Mock stored version for testing purposes.
    """
    if version is None and previous_version is not None:
        try:
            version = str(int(previous_version) + 1)  # increment version by 1
        except ValueError:
            raise ValueError(
                "Cannot auto-increment non-integer castable version, please provide"
                " manually"
            )
    return version


def get_ids_from_old_version_of_file(
    is_new_version_of: File,
    version: Optional[str],
) -> Tuple[str, str, str]:
    """{}"""
    msg = ""
    if is_new_version_of.version is None:
        previous_version = "1"
        msg = "setting version of old file to '1'"
    else:
        previous_version = is_new_version_of.version
    version = set_version(version, previous_version)
    if is_new_version_of.initial_version_id is None:
        initial_version_id = is_new_version_of.id
    else:
        initial_version_id = is_new_version_of.initial_version_id
    new_file_id = init_id(
        provisional_id=is_new_version_of.id,
        initial_version_id=initial_version_id,
        version=version,
    )
    # the following covers the edge case where the old file was unversioned
    if is_new_version_of.version is None:
        is_new_version_of.version = previous_version
        is_new_version_of.save()
        if msg != "":
            msg += (
                f"& of new file to '{version}' (initial_version_id ="
                f" '{initial_version_id}')"
            )
    return new_file_id, initial_version_id, version  # type: ignore


def __init__(file: File, *args, **kwargs):
    # Below checks for the Django-internal call in from_db()
    # it'd be better if we could avoid this, but not being able to create a File
    # from data with the default constructor renders the central class of the API
    # essentially useless
    # The danger below is not that a user might pass as many args (12 of it), but rather
    # that at some point the Django API might change; on the other hand, this
    # condition of for calling the constructor based on kwargs should always
    # stay robust
    if len(args) == len(file._meta.concrete_fields):
        super(File, file).__init__(*args, **kwargs)
        return None
    # now we proceed with the user-facing constructor
    if len(args) > 1:
        raise ValueError("Only one non-keyword arg allowed: data")
    data: Union[PathLike, DataLike] = kwargs.pop("data") if len(args) == 0 else args[0]
    key: Optional[str] = kwargs.pop("key") if "key" in kwargs else None
    run: Optional[Run] = kwargs.pop("run") if "run" in kwargs else None
    description: Optional[str] = (
        kwargs.pop("description") if "description" in kwargs else None
    )
    is_new_version_of: Optional[File] = (
        kwargs.pop("is_new_version_of") if "is_new_version_of" in kwargs else None
    )
    initial_version_id: Optional[str] = (
        kwargs.pop("initial_version_id") if "initial_version_id" in kwargs else None
    )
    version: Optional[str] = kwargs.pop("version") if "version" in kwargs else None
    format = kwargs.pop("format") if "format" in kwargs else None
    log_hint = kwargs.pop("log_hint") if "log_hint" in kwargs else True
    skip_check_exists = (
        kwargs.pop("skip_check_exists") if "skip_check_exists" in kwargs else False
    )

    if not len(kwargs) == 0:
        raise ValueError("Only data, key, run, description can be passed.")

    if is_new_version_of is None:
        provisional_id = init_id(version=version)
    else:
        if not isinstance(is_new_version_of, File):
            raise TypeError("is_new_version_of has to be of type ln.File")
        provisional_id, initial_version_id, version = get_ids_from_old_version_of_file(
            is_new_version_of, version
        )
        if description is None:
            description = is_new_version_of.description

    if version is not None:
        if initial_version_id is None:
            logger.info(
                "initializing versioning for this file! create future versions of it"
                " using ln.File(..., is_new_version_of=old_file)"
            )
    kwargs_or_file, privates = get_file_kwargs_from_data(
        data=data,
        key=key,
        run=run,
        format=format,
        provisional_id=provisional_id,
        skip_check_exists=skip_check_exists,
    )

    # an object with the same hash already exists
    if isinstance(kwargs_or_file, File):
        existing_file = kwargs_or_file
        # this is the way Django instantiates from the DB internally
        # https://github.com/django/django/blob/549d6ffeb6d626b023acc40c3bb2093b4b25b3d6/django/db/models/base.py#LL488C1-L491C51
        new_args = [
            getattr(existing_file, field.attname)
            for field in file._meta.concrete_fields
        ]
        super(File, file).__init__(*new_args)
        file._state.adding = False
        file._state.db = "default"
        return None
    else:
        kwargs = kwargs_or_file

    if isinstance(data, pd.DataFrame):
        if log_hint:
            logger.hint(
                "file is a dataframe, consider using File.from_df() to link column"
                " names as features"
            )
        kwargs["accessor"] = "DataFrame"
    elif data_is_anndata(data):
        if log_hint:
            logger.hint(
                "file is AnnDataLike, consider using File.from_anndata() to link"
                " var_names and obs.columns as features"
            )
        kwargs["accessor"] = "AnnData"
    elif data_is_mudata(data):
        kwargs["accessor"] = "MuData"

    kwargs["id"] = provisional_id
    kwargs["initial_version_id"] = initial_version_id
    kwargs["version"] = version
    kwargs["description"] = description
    # this check needs to come down here because key might be populated from an
    # existing file path during get_file_kwargs_from_data()
    if (
        kwargs["key"] is None
        and kwargs["description"] is None
        and kwargs["run"] is None
    ):
        raise ValueError("Pass one of key, run or description as a parameter")

    # transform cannot be directly passed, just via run
    # it's directly stored in the file table to avoid another join
    # mediate by the run table
    if kwargs["run"] is not None:
        if kwargs["run"].transform_id is not None:
            kwargs["transform_id"] = kwargs["run"].transform_id
        else:
            # accessing the relationship should always be possible if
            # the above if clause was false as then, we should have a fresh
            # Transform object that is not queried from the DB
            assert kwargs["run"].transform is not None
            kwargs["transform"] = kwargs["run"].transform

    if data is not None:
        file._local_filepath = privates["local_filepath"]
        file._cloud_filepath = privates["cloud_filepath"]
        file._memory_rep = privates["memory_rep"]
        file._to_store = not privates["check_path_in_storage"]

    super(File, file).__init__(**kwargs)


@classmethod  # type: ignore
@doc_args(File.from_df.__doc__)
def from_df(
    cls,
    df: "pd.DataFrame",
    columns_ref: FieldAttr = Feature.name,
    key: Optional[str] = None,
    description: Optional[str] = None,
    run: Optional[Run] = None,
) -> "File":
    """{}"""
    file = File(data=df, key=key, run=run, description=description, log_hint=False)
    feature_set = FeatureSet.from_df(df)
    if feature_set is not None:
        file._feature_sets = {"columns": feature_set}
    else:
        file._feature_sets = {}
    return file


@classmethod  # type: ignore
@doc_args(File.from_anndata.__doc__)
def from_anndata(
    cls,
    adata: "AnnDataLike",
    var_ref: Optional[FieldAttr],
    obs_columns_ref: Optional[FieldAttr] = Feature.name,
    key: Optional[str] = None,
    description: Optional[str] = None,
    run: Optional[Run] = None,
) -> "File":
    """{}"""
    file = File(data=adata, key=key, run=run, description=description, log_hint=False)
    data_parse = adata
    if not isinstance(adata, AnnData):  # is a path
        filepath = adata if isinstance(adata, (Path, UPath)) else _str_to_path(adata)
        if isinstance(filepath, UPath):
            from lamindb.dev.storage._backed_access import backed_access

            data_parse = backed_access(filepath)
        else:
            data_parse = ad.read(filepath, backed="r")
        type = "float"
    else:
        type = convert_numpy_dtype_to_lamin_feature_type(adata.X.dtype)
    feature_sets = {}
    logger.info("parsing feature names of X stored in slot 'var'")
    logger.indent = "   "
    feature_set_var = FeatureSet.from_values(
        data_parse.var.index,
        var_ref,
        type=type,
    )

    if feature_set_var is not None:
        feature_sets["var"] = feature_set_var
        logger.save(f"linked: {feature_set_var}")
    logger.indent = ""
    if len(data_parse.obs.columns) > 0:
        logger.info("parsing feature names of slot 'obs'")
        logger.indent = "   "
        feature_set_obs = FeatureSet.from_df(data_parse.obs, modality="meta")
        if feature_set_obs is not None:
            feature_sets["obs"] = feature_set_obs
            logger.save(f"linked: {feature_set_obs}")
        logger.indent = ""
    file._feature_sets = feature_sets
    return file


@classmethod  # type: ignore
@doc_args(File.from_dir.__doc__)
def from_dir(
    cls,
    path: PathLike,
    key: Optional[str] = None,
    *,
    run: Optional[Run] = None,
) -> List["File"]:
    """{}"""
    folderpath = path if isinstance(path, (Path, UPath)) else _str_to_path(path)
    storage, use_existing_storage = process_pathlike(folderpath)
    folder_key_path: Union[PurePath, Path]
    if key is None:
        if not use_existing_storage:
            logger.warning(
                "folder is outside existing storage location, will copy files from"
                f" {path} to {storage.root}/{folderpath.name}"
            )
            folder_key_path = Path(folderpath.name)
        else:
            # maintain the hierachy within an existing storage location
            folder_key_path = get_relative_path_to_directory(
                folderpath, storage.root_as_path()
            )
    else:
        folder_key_path = Path(key)

    # always sanitize by stripping a trailing slash
    folder_key = folder_key_path.as_posix().rstrip("/")

    # TODO: UPath doesn't list the first level files and dirs with "*"
    pattern = "" if isinstance(folderpath, UPath) else "*"

    # silence fine-grained logging
    verbosity = settings.verbosity
    settings.verbosity = 1  # just warnings
    files = []
    for filepath in folderpath.rglob(pattern):
        if filepath.is_file():
            relative_path = get_relative_path_to_directory(filepath, folderpath)
            file_key = folder_key + "/" + relative_path.as_posix()
            # if creating from rglob, we don't need to check for existence
            file = File(filepath, run=run, key=file_key, skip_check_exists=True)
            files.append(file)
    settings.verbosity = verbosity

    # run sanity check on hashes
    hashes = [file.hash for file in files if file.hash is not None]
    ids = [file.id for file in files]
    if len(set(hashes)) != len(hashes):
        # consider exact duplicates (same id, same hash)
        if len(set(ids)) == len(set(hashes)):
            logger.warning("dropping duplicate records in list of file records")
            files = list(set(files))
        # consider false duplicates (different id, same hash)
        else:
            seen_hashes = set()
            non_unique_files = [
                file
                for file in files
                if file.hash in seen_hashes or seen_hashes.add(file.hash)  # type: ignore  # noqa
            ]
            display_non_unique = "\n    ".join(f"{file}" for file in non_unique_files)
            logger.warning(
                "there are different file ids with the same hashes, dropping"
                f" {len(non_unique_files)} duplicates out of {len(files)} files:\n   "
                f" {display_non_unique}"
            )
            files = [
                file for file in files if file not in set(non_unique_files)
            ]  # noqa
    logger.success(
        f"created {len(files)} files from directory using storage"
        f" {storage.root} and key = {folder_key}/"
    )
    return files


def replace(
    self,
    data: Union[PathLike, DataLike],
    run: Optional[Run] = None,
    format: Optional[str] = None,
) -> None:
    kwargs, privates = get_file_kwargs_from_data(
        provisional_id=self.id,
        data=data,
        key=self.key,
        run=run,
        format=format,
    )
    if self.key is not None:
        key_path = PurePosixPath(self.key)
        new_filename = f"{key_path.stem}{kwargs['suffix']}"
        # the following will only be true if the suffix changes!
        if key_path.name != new_filename:
            self._clear_storagekey = self.key
            self.key = str(key_path.with_name(new_filename))
            logger.warning(
                f"replacing the file will replace key '{key_path}' with '{self.key}'"
                f" and delete '{key_path}' upon `save()`"
            )
    else:
        self.key = kwargs["key"]
        old_storage = auto_storage_key_from_file(self)
        new_storage = (
            self.key if self.key is not None else f"{self.id}{kwargs['suffix']}"
        )
        if old_storage != new_storage:
            self._clear_storagekey = old_storage

    self.suffix = kwargs["suffix"]
    self.size = kwargs["size"]
    self.hash = kwargs["hash"]
    self.run = kwargs["run"]
    self._local_filepath = privates["local_filepath"]
    self._cloud_filepath = privates["cloud_filepath"]
    self._memory_rep = privates["memory_rep"]
    self._to_store = not privates[
        "check_path_in_storage"
    ]  # no need to upload if new file is already in storage


def backed(
    self, is_run_input: Optional[bool] = None
) -> Union["AnnDataAccessor", "BackedAccessor"]:
    suffixes = (".h5", ".hdf5", ".h5ad", ".zrad", ".zarr")
    if self.suffix not in suffixes:
        raise ValueError(
            "File should have a zarr or h5 object as the underlying data, please use"
            " one of the following suffixes for the object name:"
            f" {', '.join(suffixes)}."
        )

    from lamindb.dev.storage._backed_access import backed_access

    _track_run_input(self, is_run_input)

    filepath = filepath_from_file(self)
    # consider the case where an object is already locally cached
    localpath = setup_settings.instance.storage.cloud_to_local_no_update(filepath)
    if localpath.exists():
        return backed_access(localpath)
    else:
        return backed_access(filepath)


def _track_run_input(file: File, is_run_input: Optional[bool] = None):
    track_run_input = False
    if is_run_input is None:
        # we need a global run context for this to work
        if run_context.run is not None:
            # avoid cycles (a file is both input and output)
            if file.run != run_context.run:
                if settings.track_run_inputs:
                    transform_note = ""
                    if file.transform is not None:
                        transform_note = (
                            f", adding parent transform {file.transform.id}"
                        )
                    logger.info(
                        f"adding file {file.id} as input for run"
                        f" {run_context.run.id}{transform_note}"
                    )
                    track_run_input = True
                else:
                    logger.hint(
                        "track this file as a run input by passing `is_run_input=True`"
                    )
        else:
            if settings.track_run_inputs:
                logger.hint(
                    "you can auto-track this file as a run input by calling"
                    " `ln.track()`"
                )
    else:
        track_run_input = is_run_input
    if track_run_input:
        if run_context.run is None:
            raise ValueError(
                "No global run context set. Call ln.context.track() or link input to a"
                " run object via `run.input_files.append(file)`"
            )
        # avoid adding the same run twice
        # avoid cycles (a file is both input and output)
        if not file.input_of.contains(run_context.run) and file.run != run_context.run:
            run_context.run.save()
            file.input_of.add(run_context.run)
            run_context.run.transform.parents.add(file.transform)


def load(self, is_run_input: Optional[bool] = None, stream: bool = False) -> DataLike:
    _track_run_input(self, is_run_input)
    if hasattr(self, "_memory_rep") and self._memory_rep is not None:
        return self._memory_rep
    return load_to_memory(filepath_from_file(self), stream=stream)


def stage(self, is_run_input: Optional[bool] = None) -> Path:
    if self.suffix in (".zrad", ".zarr"):
        raise RuntimeError("zarr object can't be staged, please use load() or stream()")
    _track_run_input(self, is_run_input)

    filepath = filepath_from_file(self)
    cb = ProgressCallback("downloading")
    return setup_settings.instance.storage.cloud_to_local(filepath, callback=cb)


def delete(self, storage: Optional[bool] = None) -> None:
    if storage is None:
        response = input(f"Are you sure you want to delete {self} from storage? (y/n)")
        delete_in_storage = response == "y"
    else:
        delete_in_storage = storage

    # need to grab file path before deletion
    filepath = self.path
    # only delete in storage if DB delete is successful
    # DB delete might error because of a foreign key constraint violated etc.
    self._delete_skip_storage()
    # we don't yet have any way to bring back the deleted metadata record
    # in case the storage deletion fails - this is important for ACID down the road
    if delete_in_storage:
        delete_storage(filepath)
        logger.success(f"deleted stored object {colors.yellow(f'{filepath}')}")


def _delete_skip_storage(file, *args, **kwargs) -> None:
    super(File, file).delete(*args, **kwargs)


def save(self, *args, **kwargs) -> None:
    self._save_skip_storage(*args, **kwargs)
    from lamindb._save import check_and_attempt_clearing, check_and_attempt_upload

    exception = check_and_attempt_upload(self)
    if exception is not None:
        self._delete_skip_storage()
        raise RuntimeError(exception)
    exception = check_and_attempt_clearing(self)
    if exception is not None:
        raise RuntimeError(exception)


def _save_skip_storage(file, *args, **kwargs) -> None:
    if file.transform is not None:
        file.transform.save()
    if file.run is not None:
        file.run.save()
    if hasattr(file, "_feature_sets"):
        for feature_set in file._feature_sets.values():
            feature_set.save()
        s = "s" if len(file._feature_sets) > 1 else ""
        logger.save(
            f"saved {len(file._feature_sets)} feature set{s} for slot{s}:"
            f" {list(file._feature_sets.keys())}"
        )
    super(File, file).save(*args, **kwargs)
    if hasattr(file, "_feature_sets"):
        links = []
        for slot, feature_set in file._feature_sets.items():
            links.append(
                File.feature_sets.through(
                    file_id=file.id, feature_set_id=feature_set.id, slot=slot
                )
            )

        from lamindb._save import bulk_create

        bulk_create(links)


@property  # type: ignore
@doc_args(File.path.__doc__)
def path(self) -> Union[Path, UPath]:
    """{}"""
    return filepath_from_file(self)


# adapted from: https://stackoverflow.com/questions/9727673
@classmethod  # type: ignore
@doc_args(File.tree.__doc__)
def tree(
    cls: File,
    path: Optional[PathLike] = None,
    *,
    level: int = -1,
    limit_to_directories: bool = False,
    length_limit: int = 1000,
) -> None:
    """{}"""
    space = "    "
    branch = "│   "
    tee = "├── "
    last = "└── "

    if path is None:
        dir_path = settings.storage
    else:
        dir_path = path if isinstance(path, (Path, UPath)) else _str_to_path(path)
    n_files = 0
    n_directories = 0

    # by default only including registered files
    # need a flag and a proper implementation
    registered_paths: Set[Any] = set()
    registered_dirs: Set[Any] = set()
    if path is None:
        registered_paths = {
            file.path for file in cls.filter(storage_id=setup_settings.storage.id).all()
        }
        registered_dirs = {d for p in registered_paths for d in p.parents}

    def inner(dir_path: Union[Path, UPath], prefix: str = "", level=-1):
        nonlocal n_files, n_directories
        if not level:
            return  # 0, stop iterating
        stripped_dir_path = dir_path.as_posix().rstrip("/")
        # do not iterate through zarr directories
        if stripped_dir_path.endswith((".zarr", ".zrad")):
            return
        # this is needed so that the passed folder is not listed
        contents = [
            i
            for i in dir_path.iterdir()
            if i.as_posix().rstrip("/") != stripped_dir_path
        ]
        if limit_to_directories:
            contents = [d for d in contents if d.is_dir()]
        pointers = [tee] * (len(contents) - 1) + [last]
        for pointer, path in zip(pointers, contents):
            if path.is_dir():
                if registered_dirs and path not in registered_dirs:
                    continue
                yield prefix + pointer + path.name
                n_directories += 1
                extension = branch if pointer == tee else space
                yield from inner(path, prefix=prefix + extension, level=level - 1)
            elif not limit_to_directories:
                if registered_paths and path not in registered_paths:
                    continue
                yield prefix + pointer + path.name
                n_files += 1

    folder_tree = ""
    iterator = inner(dir_path, level=level)
    for line in islice(iterator, length_limit):
        folder_tree += f"\n{line}"
    if next(iterator, None):
        folder_tree += f"... length_limit, {length_limit}, reached, counted:"
    directory_info = "directory" if n_directories == 1 else "directories"
    print(
        f"{dir_path.name} ({n_directories} sub-{directory_info} & {n_files} files):"
        f" {folder_tree}"
    )


def inherit_relations(self, file: File, fields: Optional[List[str]] = None):
    """Inherit many-to-many relationships from another file.

    Examples:
        >>> file1 = ln.File(pd.DataFrame(index=[0,1]))
        >>> file1.save()
        >>> file2 = ln.File(pd.DataFrame(index=[2,3]))
        >>> file2.save()
        >>> ln.save(ln.Label.from_values(["Label1", "Label2", "Label3"], field="name"))
        >>> labels = ln.Label.filter(name__icontains = "label").all()
        >>> file1.labels.set(labels)
        >>> file2.inherit_relations(file1, ["labels"])
        💬 Inheriting 1 field: ['labels']
        >>> file2.labels.list("name")
        ['Label1', 'Label2', 'Label3']
    """
    if fields is None:
        # fields in the model definition
        related_names = [i.name for i in file._meta.many_to_many]
        # fields back linked
        related_names += [i.related_name for i in file._meta.related_objects]
    else:
        related_names = []
        for field in fields:
            if hasattr(file, field):
                related_names.append(field)
            else:
                raise KeyError(f"No many-to-many relationship is found with '{field}'")

    if None in related_names:
        related_names.remove(None)

    inherit_names = [
        related_name
        for related_name in related_names
        if related_name is not None
        if file.__getattribute__(related_name).exists()
    ]

    s = "s" if len(inherit_names) > 1 else ""
    logger.info(f"inheriting {len(inherit_names)} field{s}: {inherit_names}")
    for related_name in inherit_names:
        self.__getattribute__(related_name).set(
            file.__getattribute__(related_name).all()
        )


METHOD_NAMES = [
    "__init__",
    "from_anndata",
    "from_df",
    "backed",
    "stage",
    "load",
    "delete",
    "save",
    "replace",
    "from_dir",
    "tree",
]

if _TESTING:
    from inspect import signature

    SIGS = {
        name: signature(getattr(File, name))
        for name in METHOD_NAMES
        if name != "__init__"
    }

for name in METHOD_NAMES:
    attach_func_to_class_method(name, File, globals())

# privates currently dealt with separately
File._delete_skip_storage = _delete_skip_storage
File._save_skip_storage = _save_skip_storage
# TODO: move these to METHOD_NAMES
setattr(File, "view_lineage", view_lineage)
setattr(File, "inherit_relations", inherit_relations)
setattr(File, "path", path)



import lnschema_bionty as lb
import pandas as pd
import pytest

import lamindb as ln  # noqa


@pytest.fixture(scope="module")
def df():
    return pd.DataFrame(
        (
            ["T cell", "CL:0000084"],
            ["hepatocyte", "CL:0000182"],
            ["my new cell type", ""],
        ),
        columns=["cell_type", "cell_type_id"],
    )


def test_from_values_name(df):
    assert df["cell_type_id"].tolist() == ["CL:0000084", "CL:0000182", ""]
    result = lb.CellType.from_values(df.cell_type, "name")
    ids = [i.ontology_id for i in result]
    assert len(result) == 3
    assert set(ids) == {"CL:0000084", "CL:0000182", None}
    for r in result:
        if r.ontology_id == "CL:0000084":
            assert r.bionty_source.entity == "CellType"
        if r.ontology_id is None:
            assert r.bionty_source is None
    with pytest.raises(TypeError):
        result = lb.CellType.from_values(df.cell_type, field=lb.CellType)


def test_from_values_ontology_id(df):
    result = lb.CellType.from_values(df.cell_type_id, "ontology_id")
    names = {i.name for i in result}
    assert len(result) == 2
    assert names == {"T cell", "hepatocyte"}
    assert result[0].bionty_source.entity == "CellType"


def test_from_values_multiple_match():
    records = lb.Gene.from_values(["ABC1", "PDCD1"], lb.Gene.symbol, species="human")
    assert len(records) == 3


def test_from_values_species():
    from lnschema_bionty import Gene, settings

    settings._species = None

    with pytest.raises(AssertionError):
        Gene.from_values(["ABC1"], Gene.symbol)

    settings.species = "human"
    values = ["ABC1"]
    curated_values = Gene.bionty().standardize(values)
    print(curated_values)
    records = Gene.from_values(curated_values, Gene.symbol)
    assert records[0].ensembl_gene_id == "ENSG00000068097"

    settings.species = "mouse"
    values = ["ABC1"]
    curated_values = Gene.bionty().standardize(values)
    print(curated_values)
    records = Gene.from_values(curated_values, Gene.symbol)
    assert records[0].ensembl_gene_id == "ENSMUSG00000015243"


def test_get_or_create_records():
    labels = ln.Label.from_values(["label" + str(i) for i in range(25)], field="name")
    ln.save(labels)
    # more than 20 existing values
    ln.Label.from_values(["label" + str(i) for i in range(25)], field="name")



# Generated by Django 4.2.2 on 2023-08-06 09:46

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("lnschema_core", "0013_remove_feature_labels_orm_and_more"),
    ]

    operations = [
        migrations.RenameField(
            model_name="featureset",
            old_name="ref_field",
            new_name="registry",
        ),
    ]



import pytest

from lamindb_setup.dev._hub_core import init_instance, load_instance
from lamindb_setup.dev._hub_crud import (
    sb_select_collaborator,
    sb_select_db_user_by_instance,
    sb_select_instance_by_name,
)
from lamindb_setup.dev._hub_utils import LaminDsn


def db_name(test_instance_name):
    return f"postgresql://postgres:pwd@fakeserver.xyz:5432/{test_instance_name}"


@pytest.fixture(scope="session")
def s3_bucket_1():
    return "lndb-setup-ci"


@pytest.fixture(scope="session")
def instance_1(auth_1, instance_name_1, user_account_1, account_hub_1, s3_bucket_1):
    init_instance(
        owner=auth_1["handle"],
        name=instance_name_1,
        storage=f"s3://{s3_bucket_1}",
        db=db_name(instance_name_1),
        _access_token=auth_1["access_token"],
    )
    instance = sb_select_instance_by_name(
        account_id=user_account_1["id"],
        name=instance_name_1,
        supabase_client=account_hub_1,
    )
    return instance


def test_connection_string(account_hub_1, instance_1, instance_name_1, user_account_1):
    db_user = sb_select_db_user_by_instance(
        instance_id=instance_1["id"],
        supabase_client=account_hub_1,
    )
    assert instance_1["db_scheme"] == "postgresql"
    assert instance_1["db_host"] == "fakeserver.xyz"
    assert instance_1["db_port"] == 5432
    assert instance_1["db_database"] == instance_name_1
    assert db_user["db_user_name"] == "postgres"
    assert db_user["db_user_password"] == "pwd"

    db_collaborator = sb_select_collaborator(
        instance_id=instance_1["id"],
        account_id=user_account_1["id"],
        supabase_client=account_hub_1,
    )
    assert db_collaborator["db_user_id"] == db_user["id"]


def test_load_instance(auth_1, instance_1, account_hub_1):
    result = load_instance(
        owner=auth_1["handle"],
        name=instance_1["name"],
        _access_token=auth_1["access_token"],
    )
    db_user = sb_select_db_user_by_instance(
        instance_id=instance_1["id"],
        supabase_client=account_hub_1,
    )
    expected_dsn = LaminDsn.build(
        scheme=instance_1["db_scheme"],
        user=db_user["db_user_name"],
        password=db_user["db_user_password"],
        host=instance_1["db_host"],
        port=str(instance_1["db_port"]),
        database=instance_1["db_database"],
    )
    loaded_instance, _ = result
    assert loaded_instance["name"] == instance_1["name"]
    assert loaded_instance["db"] == expected_dsn



from pathlib import Path

import lamindb as ln


def test_settings_switch_storage():
    assert ln.settings.storage.resolve() == Path("./default_storage").resolve()
    ln.settings.storage = "s3://lamindb-ci"
    assert ln.setup.settings.storage.is_cloud
    assert ln.setup.settings.storage.root_as_str == "s3://lamindb-ci"
    # root.fs contains the underlying fsspec filesystem
    # the following is set by lamindb to True for s3 by default
    assert ln.setup.settings.storage.root.fs.cache_regions
    ln.settings.storage = "s3://lamindb-ci", dict(cache_regions=False)
    assert not ln.setup.settings.storage.root.fs.cache_regions
    assert ln.Storage.filter(root="s3://lamindb-ci").one_or_none() is not None
    # switch back to default storage
    ln.settings.storage = "./default_storage"



from typing import Dict, Iterable, List, Literal, Optional, Set, Union

import numpy as np
import pandas as pd
from django.core.exceptions import FieldDoesNotExist
from django.db.models import QuerySet
from lamin_utils import logger
from lamindb_setup.dev._docs import doc_args
from lnschema_core import Registry, SynonymsAware
from lnschema_core.types import ListLike

from lamindb.dev.utils import attach_func_to_class_method

from . import _TESTING
from ._registry import get_default_str_field
from ._validate import _filter_query_based_on_species, _has_species_field


@classmethod  # type: ignore
@doc_args(SynonymsAware.standardize.__doc__)
def standardize(
    cls,
    values: Iterable,
    *,
    return_mapper: bool = False,
    case_sensitive: bool = False,
    keep: Literal["first", "last", False] = "first",
    synonyms_field: str = "synonyms",
    field: Optional[str] = None,
    **kwargs,
) -> Union[List[str], Dict[str, str]]:
    """{}"""
    return _standardize(
        cls=cls,
        values=values,
        return_mapper=return_mapper,
        case_sensitive=case_sensitive,
        keep=keep,
        synonyms_field=synonyms_field,
        field=field,
        **kwargs,
    )


def set_abbr(self, value: str):
    if hasattr(self, "name") and value == self.name:
        pass
    else:
        try:
            self.add_synonym(value, save=False)
        except NotImplementedError:
            pass
    if not hasattr(self, "abbr"):
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute 'abbr'"
        )
    self.abbr = value
    if not self._state.adding:
        self.save()


def add_synonym(
    self,
    synonym: Union[str, ListLike],
    force: bool = False,
    save: Optional[bool] = None,
):
    _check_synonyms_field_exist(self)
    _add_or_remove_synonyms(
        synonym=synonym, record=self, force=force, action="add", save=save
    )


def remove_synonym(self, synonym: Union[str, ListLike]):
    _check_synonyms_field_exist(self)
    _add_or_remove_synonyms(synonym=synonym, record=self, action="remove")


def _add_or_remove_synonyms(
    synonym: Union[str, Iterable],
    record: Registry,
    action: Literal["add", "remove"],
    force: bool = False,
    save: Optional[bool] = None,
):
    """Add or remove synonyms."""

    def check_synonyms_in_all_records(synonyms: Set[str], record: Registry):
        """Errors if input synonym is associated with other records in the DB."""
        import pandas as pd
        from IPython.display import display

        syns_all = (
            record.__class__.objects.exclude(synonyms="").exclude(synonyms=None).all()
        )
        if len(syns_all) == 0:
            return
        df = pd.DataFrame(syns_all.values())
        df["synonyms"] = df["synonyms"].str.split("|")
        df = df.explode("synonyms")
        matches_df = df[(df["synonyms"].isin(synonyms)) & (df["id"] != record.id)]
        if matches_df.shape[0] > 0:
            records_df = pd.DataFrame(syns_all.filter(id__in=matches_df["id"]).values())
            logger.error(
                f"input synonyms {matches_df['synonyms'].unique()} already associated"
                " with the following records:\n"
            )
            display(records_df)
            raise SystemExit(AssertionError)

    # passed synonyms
    # nothing happens when passing an empty string or list
    if isinstance(synonym, str):
        if len(synonym) == 0:
            return
        syn_new_set = set([synonym])
    else:
        if synonym == [""]:
            return
        syn_new_set = set(synonym)
    # nothing happens when passing an empty string or list
    if len(syn_new_set) == 0:
        return
    # because we use | as the separator
    if any(["|" in i for i in syn_new_set]):
        raise AssertionError("a synonym can't contain '|'!")

    # existing synonyms
    syns_exist = record.synonyms
    if syns_exist is None or len(syns_exist) == 0:
        syns_exist_set = set()
    else:
        syns_exist_set = set(syns_exist.split("|"))

    if action == "add":
        if not force:
            check_synonyms_in_all_records(syn_new_set, record)
        syns_exist_set.update(syn_new_set)
    elif action == "remove":
        syns_exist_set = syns_exist_set.difference(syn_new_set)

    if len(syns_exist_set) == 0:
        syns_str = None
    else:
        syns_str = "|".join(syns_exist_set)

    record.synonyms = syns_str

    if save is None:
        # if record is already in DB, save the changes to DB
        save = not record._state.adding
    if save:
        record.save()


def _check_synonyms_field_exist(record: Registry):
    try:
        record.__getattribute__("synonyms")
    except AttributeError:
        raise NotImplementedError(
            f"No synonyms field found in table {record.__class__.__name__}!"
        )


def _standardize(
    cls,
    values: Iterable,
    *,
    return_mapper: bool = False,
    case_sensitive: bool = False,
    keep: Literal["first", "last", False] = "first",
    synonyms_field: str = "synonyms",
    field: Optional[str] = None,
    **kwargs,
) -> Union[List[str], Dict[str, str]]:
    """{}"""
    from lamin_utils._map_synonyms import map_synonyms

    return_str = False
    if isinstance(values, str):
        values = [values]
        return_str = True
    if field is None:
        field = get_default_str_field(cls)
    if not isinstance(field, str):
        field = field.field.name

    orm = cls.model if isinstance(cls, QuerySet) else cls

    species = kwargs.get("species")
    if _has_species_field(orm):
        # here, we can safely import lnschema_bionty
        from lnschema_bionty._bionty import create_or_get_species_record

        species_record = create_or_get_species_record(
            species=species,
            orm=orm.model if isinstance(orm, QuerySet) else orm,
        )
        species = species_record.name if species_record is not None else species_record

    try:
        orm._meta.get_field(synonyms_field)
        df = _filter_query_based_on_species(orm=orm, species=species)
    except FieldDoesNotExist:
        df = pd.DataFrame()

    _kwargs = dict(
        field=field,
        case_sensitive=case_sensitive,
        keep=keep,
        synonyms_field=synonyms_field,
    )
    # standardized names from the DB
    std_names_db = map_synonyms(
        df=df,
        identifiers=values,
        return_mapper=return_mapper,
        **_kwargs,
    )

    def _return(result: List, mapper: Dict, return_mapper: bool, return_str: bool):
        if return_mapper:
            return mapper
        else:
            if return_str and len(result) == 1:
                return result[0]
            return result

    # map synonyms in Bionty
    if orm.__get_schema_name__() == "bionty":
        mapper = {}
        if return_mapper:
            mapper = std_names_db
            std_names_db = map_synonyms(
                df=df, identifiers=values, return_mapper=False, **_kwargs
            )

        val_res = orm.validate(std_names_db, field=field, mute=True, species=species)
        if all(val_res):
            return _return(
                result=std_names_db,
                mapper=mapper,
                return_mapper=return_mapper,
                return_str=return_str,
            )

        nonval = np.array(std_names_db)[~val_res]
        std_names_bt_mapper = orm.bionty(species=species).standardize(
            nonval, return_mapper=True, **_kwargs
        )

        if len(std_names_bt_mapper) > 0:
            s = "" if len(std_names_bt_mapper) == 1 else "s"
            warn_msg = (
                f"found {len(std_names_bt_mapper)} synonym{s} in Bionty:"
                f" {list(std_names_bt_mapper.keys())}"
            )
            warn_msg += (
                "\n   please add corresponding records via"
                f" `.from_values({list(std_names_bt_mapper.values())})`"
            )
            logger.warning(warn_msg)

        mapper.update(std_names_bt_mapper)
        result = pd.Series(std_names_db).replace(std_names_bt_mapper).tolist()
        return _return(
            result=result,
            mapper=mapper,
            return_mapper=return_mapper,
            return_str=return_str,
        )

    else:
        return _return(
            result=std_names_db,
            mapper=std_names_db,
            return_mapper=return_mapper,
            return_str=return_str,
        )


METHOD_NAMES = ["standardize", "add_synonym", "remove_synonym", "set_abbr"]

if _TESTING:  # type: ignore
    from inspect import signature

    SIGS = {
        name: signature(getattr(SynonymsAware, name))
        for name in METHOD_NAMES
        if not name.startswith("__")
    }

for name in METHOD_NAMES:
    attach_func_to_class_method(name, SynonymsAware, globals())



# Generated by Django 4.2.2 on 2023-07-24 16:50

import django.db.models.deletion
from django.db import migrations, models

import lnschema_core.ids


class Migration(migrations.Migration):
    dependencies = [
        ("lnschema_core", "0011_label_remove_tag_created_by_remove_tag_parents_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="label",
            name="ref_id",
        ),
        migrations.RemoveField(
            model_name="label",
            name="ref_orm",
        ),
        migrations.RemoveField(
            model_name="label",
            name="ref_schema",
        ),
        migrations.AddField(
            model_name="feature",
            name="labels_orm",
            field=models.CharField(db_index=True, default=None, max_length=40, null=True),
        ),
        migrations.AddField(
            model_name="feature",
            name="labels_schema",
            field=models.CharField(db_index=True, default=None, max_length=40, null=True),
        ),
        migrations.CreateModel(
            name="FileFeatureSet",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("slot", models.CharField(default=None, max_length=40, null=True)),
                ("feature_set", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="lnschema_core.featureset")),
                ("file", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="lnschema_core.file")),
            ],
            options={
                "unique_together": {("file", "feature_set")},
            },
        ),
        migrations.CreateModel(
            name="DatasetFeatureSet",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("slot", models.CharField(default=None, max_length=50, null=True)),
                ("dataset", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="lnschema_core.dataset")),
                ("feature_set", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="lnschema_core.featureset")),
            ],
            options={
                "unique_together": {("dataset", "feature_set")},
            },
        ),
        migrations.RunSQL("CREATE TABLE lnschema_core_filefeatureset_tmp (id BIGINT, file_id TEXT, feature_set_id TEXT)"),
        migrations.RunSQL("INSERT INTO lnschema_core_filefeatureset_tmp (id, file_id, feature_set_id) SELECT id, file_id, featureset_id from lnschema_core_file_feature_sets"),
        migrations.RemoveField(
            model_name="file",
            name="feature_sets",
        ),
        migrations.RemoveField(
            model_name="dataset",
            name="feature_sets",
        ),
        migrations.AddField(
            model_name="dataset",
            name="feature_sets",
            field=models.ManyToManyField(related_name="datasets", through="lnschema_core.DatasetFeatureSet", to="lnschema_core.featureset"),
        ),
        migrations.AddField(
            model_name="file",
            name="feature_sets",
            field=models.ManyToManyField(related_name="files", through="lnschema_core.FileFeatureSet", to="lnschema_core.featureset"),
        ),
        migrations.RunSQL("INSERT INTO lnschema_core_filefeatureset (id, file_id, feature_set_id) SELECT id, file_id, feature_set_id from lnschema_core_filefeatureset_tmp"),
        migrations.RunSQL("DROP TABLE lnschema_core_filefeatureset_tmp"),
        migrations.AddField(
            model_name="file",
            name="accessor",
            field=models.CharField(db_index=True, default=None, max_length=64, null=True),
        ),
        migrations.RemoveField(
            model_name="featureset",
            name="readout",
        ),
        migrations.CreateModel(
            name="Modality",
            fields=[
                ("id", models.CharField(default=lnschema_core.ids.base62_8, max_length=8, primary_key=True, serialize=False)),
                ("name", models.CharField(db_index=True, max_length=256)),
                ("ontology_id", models.CharField(db_index=True, default=None, max_length=32, null=True)),
                ("abbr", models.CharField(db_index=True, default=None, max_length=32, null=True, unique=True)),
                ("synonyms", models.TextField(default=None, null=True)),
                ("description", models.TextField(default=None, null=True)),
                ("molecule", models.TextField(db_index=True, default=None, null=True)),
                ("instrument", models.TextField(db_index=True, default=None, null=True)),
                ("measurement", models.TextField(db_index=True, default=None, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id, on_delete=django.db.models.deletion.PROTECT, related_name="created_modalities", to="lnschema_core.user"
                    ),
                ),
                ("parents", models.ManyToManyField(related_name="children", to="lnschema_core.modality")),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="featureset",
            name="modality",
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.PROTECT, to="lnschema_core.modality"),
        ),
        migrations.AlterField(
            model_name="featureset",
            name="created_by",
            field=models.ForeignKey(
                default=lnschema_core.users.current_user_id, on_delete=django.db.models.deletion.PROTECT, related_name="created_feature_sets", to="lnschema_core.user"
            ),
        ),
        migrations.AddField(
            model_name="featureset",
            name="hash",
            field=models.CharField(db_index=True, default=None, max_length=20, null=True),
        ),
    ]



from typing import List, Optional

from lamindb_setup.dev._docs import doc_args
from lnschema_core import Label
from lnschema_core.types import ListLike

from lamindb.dev.utils import attach_func_to_class_method

from . import _TESTING
from ._from_values import get_or_create_records


def __init__(self, *args, **kwargs):
    if len(args) == len(self._meta.concrete_fields):
        super(Label, self).__init__(*args, **kwargs)
        return None
    # now we proceed with the user-facing constructor
    if len(args) > 0:
        raise ValueError("Only one non-keyword arg allowed")
    name: Optional[str] = kwargs.pop("name") if "name" in kwargs else None
    description: Optional[str] = (
        kwargs.pop("description") if "description" in kwargs else None
    )
    if len(kwargs) > 0:
        raise ValueError("Only name & description are valid keyword arguments")
    super(Label, self).__init__(
        name=name,
        description=description,
    )


@classmethod  # type:ignore
@doc_args(Label.from_values.__doc__)
def from_values(cls, values: ListLike, **kwargs) -> List["Label"]:
    """{}"""
    records = get_or_create_records(
        iterable=values,
        field=Label.name,
    )
    return records


METHOD_NAMES = [
    "__init__",
    "from_values",
]

if _TESTING:
    from inspect import signature

    SIGS = {
        name: signature(getattr(Label, name))
        for name in METHOD_NAMES
        if name != "__init__"
    }

for name in METHOD_NAMES:
    attach_func_to_class_method(name, Label, globals())



from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd
from django.db.models import QuerySet
from lamin_utils._inspect import InspectResult
from lamindb_setup.dev._docs import doc_args
from lnschema_core import Registry, ValidationAware
from lnschema_core.types import ListLike, StrField

from lamindb.dev.utils import attach_func_to_class_method

from . import _TESTING
from ._from_values import _has_species_field


@classmethod  # type: ignore
@doc_args(ValidationAware.inspect.__doc__)
def inspect(
    cls,
    values: ListLike,
    field: StrField,
    *,
    mute: bool = False,
    **kwargs,
) -> InspectResult:
    """{}"""
    return _inspect(
        cls=cls,
        values=values,
        field=field,
        mute=mute,
        **kwargs,
    )


@classmethod  # type: ignore
@doc_args(ValidationAware.validate.__doc__)
def validate(
    cls, values: ListLike, field: StrField, *, mute: bool = False, **kwargs
) -> np.ndarray:
    """{}"""
    return _validate(cls=cls, values=values, field=field, mute=mute, **kwargs)


def _inspect(
    cls,
    values: ListLike,
    field: StrField,
    *,
    mute: bool = False,
    **kwargs,
) -> Union["pd.DataFrame", Dict[str, List[str]]]:
    """{}"""
    from lamin_utils._inspect import inspect

    if isinstance(values, str):
        values = [values]
    if not isinstance(field, str):
        field = field.field.name

    orm = cls.model if isinstance(cls, QuerySet) else cls

    return inspect(
        df=_filter_query_based_on_species(orm=orm, species=kwargs.get("species")),
        identifiers=values,
        field=str(field),
        mute=mute,
        **kwargs,
    )


def _validate(
    cls, values: ListLike, field: StrField, *, mute: bool = False, **kwargs
) -> np.ndarray:
    """{}"""
    from lamin_utils._inspect import validate

    if isinstance(values, str):
        values = [values]
    if not isinstance(field, str):
        field = field.field.name

    orm = cls.model if isinstance(cls, QuerySet) else cls
    field_values = pd.Series(
        _filter_query_based_on_species(
            orm=orm,
            species=kwargs.get("species"),
            values_list_field=field,
        ),
        dtype="object",
    )
    return validate(
        identifiers=values,
        field_values=field_values,
        case_sensitive=True,
        mute=mute,
        field=field,
        **kwargs,
    )


def _filter_query_based_on_species(
    orm: Union[Registry, QuerySet],
    species: Optional[Union[str, Registry]] = None,
    values_list_field: Optional[str] = None,
):
    import pandas as pd

    if values_list_field is None:
        records = orm.all() if isinstance(orm, QuerySet) else orm.objects.all()
    else:
        records = orm if isinstance(orm, QuerySet) else orm.objects
    if _has_species_field(orm):
        # here, we can safely import lnschema_bionty
        from lnschema_bionty._bionty import create_or_get_species_record

        species_record = create_or_get_species_record(
            species=species, orm=orm.model if isinstance(orm, QuerySet) else orm
        )
        if species_record is not None:
            records = records.filter(species__name=species_record.name)

    if values_list_field is None:
        return pd.DataFrame.from_records(records.values())
    else:
        return records.values_list(values_list_field, flat=True)


METHOD_NAMES = [
    "validate",
    "inspect",
]

if _TESTING:  # type: ignore
    from inspect import signature

    SIGS = {
        name: signature(getattr(ValidationAware, name))
        for name in METHOD_NAMES
        if not name.startswith("__")
    }

for name in METHOD_NAMES:
    attach_func_to_class_method(name, ValidationAware, globals())



from inspect import signature

import lnschema_bionty as lb
import pandas as pd
import pytest

import lamindb as ln
from lamindb import _feature_set
from lamindb._feature_set import get_related_name, sanity_check_features

df = pd.DataFrame(
    {
        "feat1": [1, 2, 3],
        "feat2": [3.1, 4.2, 5.3],
        "feat3": ["cond1", "cond2", "cond2"],
        "feat4": ["id1", "id2", "id3"],
    }
)


def test_signatures():
    # this seems currently the easiest and most transparent
    # way to test violations of the signature equality
    # the MockORM class is needed to get inspect.signature
    # to work
    class Mock:
        pass

    # class methods
    class_methods = ["from_values", "from_df"]
    for name in class_methods:
        setattr(Mock, name, getattr(_feature_set, name))
        assert signature(getattr(Mock, name)) == _feature_set.SIGS.pop(name)
    # methods
    for name, sig in _feature_set.SIGS.items():
        assert signature(getattr(_feature_set, name)) == sig


def test_feature_set_from_values():
    gene_symbols = ["TCF7", "MYC"]
    lb.settings.species = "human"
    feature_set = ln.FeatureSet.from_values(gene_symbols, lb.Gene.symbol)
    id = feature_set.id
    assert feature_set._state.adding
    assert feature_set.type == "float"
    assert feature_set.registry == "bionty.Gene"
    feature_set.save()
    # test that the feature_set is retrieved from the database
    # in case it already exists
    feature_set = ln.FeatureSet.from_values(gene_symbols, lb.Gene.symbol)
    assert not feature_set._state.adding
    assert id == feature_set.id
    feature_set.delete()

    # edge cases
    with pytest.raises(ValueError):
        feature_set = ln.FeatureSet.from_values([])

    with pytest.raises(TypeError):
        ln.FeatureSet.from_values(["a"], field="name")
    with pytest.raises(ValueError):
        ln.FeatureSet.from_values(["name"], field=ln.Feature.name, type="rna")
    with pytest.raises(TypeError):
        ln.FeatureSet.from_values([1], field=ln.Label.name, type="rna")

    # return none if no validated features
    assert ln.FeatureSet.from_values(["name"], field=ln.Label.name, type="rna") is None


def test_feature_set_from_records():
    features = ln.Feature.from_df(df)
    feature_set = ln.FeatureSet(features)
    id = feature_set.id
    assert feature_set._state.adding
    assert feature_set.type is None
    assert feature_set.registry == "core.Feature"
    feature_set.save()
    # test that the feature_set is retrieved from the database
    # in case it already exists
    feature_set = ln.FeatureSet(features)
    assert not feature_set._state.adding
    assert id == feature_set.id
    feature_set.delete()

    # edge case
    with pytest.raises(ValueError):
        positional_arg = 1
        ln.FeatureSet(features, positional_arg)


def test_feature_set_from_df():
    feature_set = ln.FeatureSet.from_df(df)
    feature_set.save()
    assert feature_set.type is None
    for feature in feature_set.features.all():
        feature.delete()
    feature_set.delete()


def test_get_related_name():
    with pytest.raises(ValueError):
        get_related_name(ln.Transform)


def test_sanity_check_features():
    with pytest.raises(ValueError):
        sanity_check_features([])
    with pytest.raises(TypeError):
        sanity_check_features(["feature"])
    with pytest.raises(TypeError):
        sanity_check_features({"feature"})
    with pytest.raises(ValueError):
        sanity_check_features([ln.Run(), ln.Transform()])


def test_kwargs():
    with pytest.raises(ValueError):
        ln.FeatureSet(x="1", features=[])


def test_pass_modality():
    m = ln.Modality(name="rna")
    feature_set = ln.FeatureSet(modality=m, features=[ln.Feature(name="f", type="f")])
    assert feature_set.modality == m
    with pytest.raises(ValueError):
        feature_set = ln.FeatureSet(
            modality=ln.Transform(), features=[ln.Feature(name="f", type="f")]
        )



from typing import List, Optional, Union

import pandas as pd
from lamindb_setup.dev._docs import doc_args
from lnschema_core import Feature
from lnschema_core.models import Registry
from pandas.api.types import is_categorical_dtype, is_string_dtype

from lamindb.dev.utils import attach_func_to_class_method

from . import _TESTING


def convert_numpy_dtype_to_lamin_feature_type(dtype) -> str:
    orig_type = dtype.name
    # strip precision qualifiers
    type = "".join(i for i in orig_type if not i.isdigit())
    return type


# def take(n, iterable):
# from itertools import islice
#     """Return the first n items of the iterable as a list."""
#     return list(islice(iterable, n))


def __init__(self, *args, **kwargs):
    if len(args) == len(self._meta.concrete_fields):
        super(Feature, self).__init__(*args, **kwargs)
        return None
    # now we proceed with the user-facing constructor
    if len(args) != 0:
        raise ValueError("Only non-keyword args allowed")
    type: Optional[Union[type, str]] = kwargs.pop("type") if "type" in kwargs else None
    registries: Optional[List[Registry]] = (
        kwargs.pop("registries") if "registries" in kwargs else None
    )
    # cast type
    type_str = None
    if type is not None:
        type_str = type.__name__ if not isinstance(type, str) else type
    kwargs["type"] = type_str
    # cast registries
    registries_str: Optional[str] = None
    if registries is not None:
        if isinstance(registries, str):
            # TODO: add more validation
            registries_str = registries
        else:
            if not isinstance(registries, List):
                raise ValueError("registries has to be a list of Registry types")
            registries_str = ""
            for cls in registries:
                if not hasattr(cls, "__get_name_with_schema__"):
                    raise ValueError("each element of the list has to be a Registry")
                registries_str += cls.__get_name_with_schema__() + "|"
            registries_str = registries_str.rstrip("|")
    kwargs["registries"] = registries_str
    super(Feature, self).__init__(*args, **kwargs)


@classmethod  # type:ignore
@doc_args(Feature.from_df.__doc__)
def from_df(cls, df: "pd.DataFrame") -> List["Feature"]:
    """{}"""
    string_cols = [col for col in df.columns if is_string_dtype(df[col])]
    categoricals = {col: df[col] for col in df.columns if is_categorical_dtype(df[col])}
    for key in string_cols:
        c = pd.Categorical(df[key])
        if len(c.categories) < len(c):
            categoricals[key] = c

    types = {}
    # categoricals_with_unmapped_categories = {}  # type: ignore
    for name, col in df.items():
        if name in categoricals:
            types[name] = "category"
            # below is a harder feature to write, now, because it requires to
            # query the link tables between the label Registry and file or dataset
            # the original implementation fell short
            # categorical = categoricals[name]
            # if hasattr(
            #     categorical, "cat"
            # ):  # because .categories > pd2.0, .cat.categories < pd2.0
            #     categorical = categorical.cat
            # categories = categorical.categories
            # categoricals_with_unmapped_categories[name] = Label.filter(
            #     feature=name
            # ).inspect(categories, "name", logging=False)["not_mapped"]
        else:
            types[name] = convert_numpy_dtype_to_lamin_feature_type(col.dtype)

    features = Feature.from_values(df.columns, field=Feature.name, types=types)
    assert len(features) == len(df.columns)

    # if len(categoricals_with_unmapped_categories) > 0:
    #     n_max = 20
    #     categoricals_with_unmapped_categories_formatted = "\n      ".join(
    #         [
    #             (
    #                 f"{key} ({len(value)}): {', '.join(value)}"
    #                 if len(value) <= 5
    #                 else f"{key} ({len(value)}): {', '.join(value[:5])} ..."
    #             )
    #             for key, value in take(
    #                 n_max, categoricals_with_unmapped_categories.items()
    #             )
    #         ]
    #     )
    #     if len(categoricals_with_unmapped_categories) > n_max:
    #         categoricals_with_unmapped_categories_formatted += "\n      ..."
    #     categoricals_with_unmapped_categories_formatted
    #     logger.info(
    #         f"{len(categoricals_with_unmapped_categories)} features have"
    #         f" {colors.yellow('unmapped categories')}:\n     "
    #         f" {categoricals_with_unmapped_categories_formatted}"
    #     )
    return features


@doc_args(Feature.save.__doc__)
def save(self, *args, **kwargs) -> None:
    """{}"""
    super(Feature, self).save(*args, **kwargs)


METHOD_NAMES = [
    "__init__",
    "from_df",
    "save",
]

if _TESTING:
    from inspect import signature

    SIGS = {
        name: signature(getattr(Feature, name))
        for name in METHOD_NAMES
        if name != "__init__"
    }

for name in METHOD_NAMES:
    attach_func_to_class_method(name, Feature, globals())



# this tries to load a persistent instance created with
# lamin init --storage s3://lndb-setup-ci --name persistent-instance-to-test-load

import lamindb_setup as ln_setup


def test_load_persistent_instance():
    ln_setup.load("laminapp-admin/persistent-instance-to-test-load")
    ln_setup.close()



from django.db import transaction

from ._settings_instance import InstanceSettings


def write_bionty_sources(isettings: InstanceSettings):
    """Write bionty sources to BiontySource table."""
    if "bionty" in isettings.schema:
        import shutil

        from bionty.dev._handle_sources import (
            CURRENT_SOURCES,
            LAMINDB_SOURCES,
            LOCAL_SOURCES,
            parse_sources_yaml,
        )

        shutil.copy(CURRENT_SOURCES, LAMINDB_SOURCES)

        import bionty as bt
        from lnschema_bionty.models import BiontySource

        all_sources = parse_sources_yaml(LOCAL_SOURCES)
        all_sources_dict = all_sources.to_dict(orient="records")

        currently_used = (
            bt.display_currently_used_sources()
            .reset_index()
            .set_index(["entity", "species"])
        )

        all_records = []
        for kwargs in all_sources_dict:
            act = currently_used.loc[(kwargs["entity"], kwargs["species"])].to_dict()
            if (act["source"] == kwargs["source"]) and (
                act["version"] == kwargs["version"]
            ):
                kwargs["currently_used"] = True

            record = BiontySource(**kwargs)
            all_records.append(record)

        with transaction.atomic():
            for record in all_records:
                record.save()


def load_bionty_sources(isettings: InstanceSettings):
    """Write currently_used bionty sources to LAMINDB_VERSIONS_PATH in bionty."""
    if "bionty" in isettings.schema:
        from bionty.dev._handle_sources import (
            LAMINDB_SOURCES,
            parse_currently_used_sources,
        )
        from bionty.dev._io import write_yaml
        from lnschema_bionty.models import BiontySource

        try:  # this is only to deal with legacy instances
            active_records = (
                BiontySource.objects.filter(currently_used=True).all().values()
            )

            write_yaml(parse_currently_used_sources(active_records), LAMINDB_SOURCES)
        except Exception:
            pass


def delete_bionty_sources_yaml():
    """Delete LAMINDB_SOURCES_PATH in bionty."""
    try:
        from bionty.dev._handle_sources import LAMINDB_SOURCES

        LAMINDB_SOURCES.unlink(missing_ok=True)
    except ModuleNotFoundError:
        pass



"""Types.

.. autosummary::
   :toctree: .

   PathLike
   DataLike
   StrField
   ListLike
   TransformType
"""
from lnschema_core.types import (  # noqa
    DataLike,
    ListLike,
    PathLike,
    StrField,
    TransformType,
)



import lamindb_setup as ln_setup


def test_login():
    ln_setup.login("testuser1")
    assert ln_setup.settings.user.email == "testuser1@lamin.ai"
    assert ln_setup.settings.user.id == "DzTjkKse"
    assert ln_setup.settings.user.handle == "testuser1"

    import jwt

    access_token_payload = jwt.decode(
        ln_setup.settings.user.access_token,
        algorithms="HS256",
        options={"verify_signature": False},
    )
    assert access_token_payload["email"] == "testuser1@lamin.ai"



from typing import Iterable, List, Optional, Type, Union

import pandas as pd
from lamin_utils import logger
from lamindb_setup.dev._docs import doc_args
from lnschema_core import Feature, FeatureSet, Modality, Registry, ids
from lnschema_core.types import FieldAttr, ListLike

from lamindb.dev.hashing import hash_set
from lamindb.dev.utils import attach_func_to_class_method

from . import _TESTING
from ._from_values import get_or_create_records, index_iterable
from ._registry import init_self_from_db
from ._save import bulk_create


def get_related_name(features_type: Registry):
    candidates = [
        field.related_name
        for field in FeatureSet._meta.related_objects
        if field.related_model == features_type
    ]
    if not candidates:
        raise ValueError(
            f"Can't create feature sets from {features_type.__name__} because it's not"
            " related to it!\nYou need to create a link model between FeatureSet and"
            " your Registry in your custom schema.\nTo do so, add a"
            " line:\nfeature_sets = models.ManyToMany(FeatureSet,"
            " related_name='mythings')\n"
        )
    return candidates[0]


def sanity_check_features(features: List[Registry]) -> Registry:
    """Validate and return feature type."""
    if len(features) == 0:
        raise ValueError("provide list of features with at least one element")
    if not hasattr(features, "__getitem__"):
        raise TypeError("features has to be list-like")
    if not isinstance(features[0], Registry):
        raise TypeError(
            "features has to store feature records! use .from_values() otherwise"
        )
    feature_types = set([feature.__class__ for feature in features])
    if len(feature_types) > 1:
        raise ValueError("feature_set can only contain a single type")
    return next(iter(feature_types))  # return value in set of cardinality 1


def get_validated_features(
    features: List[Registry], field: FieldAttr
) -> List[Registry]:
    validated_features = []
    non_validated_features = []
    for feature in features:
        if feature._state.adding and not (
            hasattr(feature, "_from_bionty") and feature._from_bionty
        ):
            non_validated_features.append(getattr(feature, field.field.name))
        else:
            validated_features.append(feature)
    if non_validated_features:
        non_validated_features_display = ",".join(non_validated_features)
        logger.warning(
            f"ignoring non-validated features: {non_validated_features_display}"
        )
    return validated_features


def __init__(self, *args, **kwargs):
    if len(args) == len(self._meta.concrete_fields):
        super(FeatureSet, self).__init__(*args, **kwargs)
        return None
    # now we proceed with the user-facing constructor
    if len(args) > 1:
        raise ValueError("Only one non-keyword arg allowed: features")
    features: Iterable[Registry] = kwargs.pop("features") if len(args) == 0 else args[0]
    type: Optional[Union[type, str]] = kwargs.pop("type") if "type" in kwargs else None
    modality: Optional[str] = kwargs.pop("modality") if "modality" in kwargs else None
    name: Optional[str] = kwargs.pop("name") if "name" in kwargs else None
    # hash is only internally used
    hash: Optional[str] = kwargs.pop("hash") if "hash" in kwargs else None
    if len(kwargs) > 0:
        raise ValueError(
            "Only features, type, modality, name are valid keyword arguments"
        )

    # now code
    features_orm = sanity_check_features(features)
    if features_orm == Feature:
        type = None
    else:
        type = float
    n_features = len(features)
    if hash is None:
        features_hash = hash_set({feature.id for feature in features})
        feature_set = FeatureSet.filter(hash=features_hash).one_or_none()
        if feature_set is not None:
            logger.success(f"loaded: {feature_set}")
            init_self_from_db(self, feature_set)
            return None
        else:
            hash = features_hash
    self._features = (get_related_name(features_orm), features)
    if type is not None:
        type_str = type.__name__ if not isinstance(type, str) else type
    else:
        type_str = None
    if modality is not None:
        if isinstance(modality, str):
            modality_record = Modality.filter(name=modality).one_or_none()
            if modality_record is None:
                modality_record = Modality(name=modality)
                modality_record.save()
        elif isinstance(modality, Modality):
            modality_record = modality
        else:
            raise ValueError("modality needs to be string or Modality record")
    else:
        modality_record = modality
    super(FeatureSet, self).__init__(
        id=ids.base62_20(),
        name=name,
        type=type_str,
        n=n_features,
        modality=modality_record,
        registry=features_orm.__get_name_with_schema__(),
        hash=hash,
    )


@doc_args(FeatureSet.save.__doc__)
def save(self, *args, **kwargs) -> None:
    """{}"""
    super(FeatureSet, self).save(*args, **kwargs)
    if hasattr(self, "_features"):
        related_name, records = self._features
        # if values are stored in their dedicated table, we can bulk_create
        if related_name != "features":
            bulk_create(records)
        else:
            # otherwise, we currently need to save one by one
            for record in records:
                record.save()
        getattr(self, related_name).set(records)


@classmethod  # type:ignore
@doc_args(FeatureSet.from_values.__doc__)
def from_values(
    cls,
    values: ListLike,
    field: FieldAttr = Feature.name,
    type: Optional[Union[Type, str]] = None,
    name: Optional[str] = None,
    modality: Optional[str] = None,
    **kwargs,
) -> Optional["FeatureSet"]:
    """{}"""
    if not isinstance(field, FieldAttr):
        raise TypeError(
            "Argument `field` must be a Registry field, e.g., `Feature.name`"
        )
    if len(values) == 0:
        raise ValueError("Provide a list of at least one value")
    registry = field.field.model
    if registry.__name__ == "Feature":
        raise ValueError("Please use from_df() instead of from_values()")
    iterable_idx = index_iterable(values)
    if not isinstance(iterable_idx[0], (str, int)):
        raise TypeError("values should be list-like of str or int")
    from_bionty = registry.__module__.startswith("lnschema_bionty")
    features = get_or_create_records(
        iterable=iterable_idx,
        field=field,
        from_bionty=from_bionty,
        **kwargs,
    )
    validated_features = get_validated_features(features, field)
    validated_feature_ids = [feature.id for feature in validated_features]
    features_hash = hash_set(set(validated_feature_ids))
    feature_set = FeatureSet.filter(hash=features_hash).one_or_none()
    if feature_set is not None:
        logger.success(f"loaded {feature_set}")
    else:
        if type is not None:
            type_str = type.__name__ if not isinstance(type, str) else type
        else:
            type_str = None
        if validated_features:
            feature_set = FeatureSet(
                features=validated_features,
                hash=features_hash,
                name=name,
                modality=modality,
                type=type_str,
            )
        else:
            feature_set = None
    return feature_set


@classmethod  # type:ignore
@doc_args(FeatureSet.from_df.__doc__)
def from_df(
    cls,
    df: "pd.DataFrame",
    name: Optional[str] = None,
    **kwargs,
) -> Optional["FeatureSet"]:
    """{}"""
    features = Feature.from_df(df)
    validated_features = get_validated_features(features, Feature.name)
    if validated_features:
        feature_set = FeatureSet(validated_features, name=name, **kwargs)
    else:
        logger.warning("no validated features, skip creating feature set")
        feature_set = None
        # raise ValidationError("Dataframe columns contain no validated feature names")
    return feature_set


METHOD_NAMES = [
    "__init__",
    "from_values",
    "from_df",
    "save",
]

if _TESTING:
    from inspect import signature

    SIGS = {
        name: signature(getattr(FeatureSet, name))
        for name in METHOD_NAMES
        if name != "__init__"
    }

for name in METHOD_NAMES:
    attach_func_to_class_method(name, FeatureSet, globals())



import lamindb as ln


def test_track_with_multi_parents():
    parent1 = ln.Transform(name="parent 1")
    parent1.save()
    parent2 = ln.Transform(name="parent 2")
    parent2.save()
    child = ln.Transform(name="Child")
    child.save()
    child.parents.set([parent1, parent2])
    ln.track(child, reference="my address", reference_type="url")
    # unset to remove side effects
    ln.dev.run_context.run = None
    ln.dev.run_context.transform = None


def test_track_with_reference():
    transform = ln.Transform(name="test")
    ln.track(transform, reference="my address", reference_type="url")
    assert ln.dev.run_context.run.reference == "my address"
    assert ln.dev.run_context.run.reference_type == "url"
    # unset to remove side effects
    ln.dev.run_context.run = None
    ln.dev.run_context.transform = None



import os
from pathlib import Path
from typing import Any, Dict, Optional
from urllib.request import urlretrieve

from pydantic import BaseSettings, PostgresDsn
from supabase import create_client
from supabase.lib.client_options import ClientOptions


class Connector(BaseSettings):
    url: str
    key: str


class Settings(BaseSettings):
    lamin_env: str = "local"
    postgres_dsn: Optional[PostgresDsn]
    supabase_api_url: str
    supabase_anon_key: str
    supabase_service_role_key: Optional[str]
    ln_server_deploy: int = 0

    class Config:
        # This will look for an env file in the parent directory,
        # e.g. prod.env, staging.env, or local.env. Explicitly set
        # environment variables will take precendence over the ones
        # read in by file.

        env_file = (
            Path(__file__).parents[1] / f"lnhub-rest--{os.environ['LAMIN_ENV']}.env"
            if os.environ.get("LAMIN_ENV")
            else None
        )

        @classmethod
        def customise_sources(cls, init_settings, env_settings, file_secret_settings):
            # Following pattern here:
            # https://docs.pydantic.dev/usage/settings/#customise-settings-sources
            # in the absence of init and env settings, pull from S3
            return (init_settings, env_settings, lamindb_client_config_settings)


def connect_hub(client_options: ClientOptions = ClientOptions()):
    settings = Settings()

    return create_client(
        settings.supabase_api_url, settings.supabase_anon_key, client_options
    )
    # This is now handled in lamindb_client_config_settings
    # connector_file, _ = urlretrieve(
    #     "https://lamin-site-assets.s3.amazonaws.com/connector.env"
    # )
    # connector = Connector(_env_file=connector_file)
    # return create_client(connector.url, connector.key)


def connect_hub_with_auth(
    *,
    email: Optional[str] = None,
    password: Optional[str] = None,
    access_token: Optional[str] = None,
):
    hub = connect_hub()
    if access_token is None:
        if email is None or password is None:
            from .._settings_load import load_or_create_user_settings

            user_settings = load_or_create_user_settings()
            email = user_settings.email
            password = user_settings.password
        access_token = get_access_token(email=email, password=password)
    hub.postgrest.auth(access_token)
    return hub


def get_access_token(email: Optional[str] = None, password: Optional[str] = None):
    hub = connect_hub()
    try:
        auth_response = hub.auth.sign_in_with_password(
            {
                "email": email,
                "password": password,
            }
        )
        return auth_response.session.access_token
    finally:
        hub.auth.sign_out()


def lamindb_client_config_settings(settings: BaseSettings) -> Dict[str, Any]:
    if os.getenv("LAMIN_ENV") == "staging":
        connector_path = (
            "https://lamin-site-assets.s3.amazonaws.com/connector_staging.env"
        )
    else:
        connector_path = "https://lamin-site-assets.s3.amazonaws.com/connector.env"
    connector_file, _ = urlretrieve(connector_path)
    connector = Connector(_env_file=connector_file)
    return dict(
        lamin_env="client",
        supabase_api_url=connector.url,
        supabase_anon_key=connector.key,
    )


def get_lamin_site_base_url():
    if "LAMIN_ENV" in os.environ:
        if os.environ["LAMIN_ENV"] == "local":
            return "http://localhost:3000"
        elif os.environ["LAMIN_ENV"] == "staging":
            return "https://staging.lamin.ai"
    return "https://lamin.ai"



from typing import Iterable, List, Optional, Tuple, Union

import anndata as ad
import pandas as pd
from lnschema_core.models import Dataset

from . import FeatureSet, File, Run
from .dev.hashing import hash_set


def __init__(
    dataset: Dataset,
    *args,
    **kwargs,
):
    if len(args) == len(dataset._meta.concrete_fields):
        super(Dataset, dataset).__init__(*args, **kwargs)
        return None
    # now we proceed with the user-facing constructor
    if len(args) > 1:
        raise ValueError("Only one non-keyword arg allowed: data")
    data: Optional[Union[pd.DataFrame, ad.AnnData]] = None
    if "data" in kwargs or len(args) == 1:
        data = kwargs.pop("data") if len(args) == 0 else args[0]
    name: Optional[str] = kwargs.pop("name") if "name" in kwargs else None
    description: Optional[str] = (
        kwargs.pop("description") if "description" in kwargs else None
    )
    run: Optional[Run] = kwargs.pop("run") if "run" in kwargs else None
    assert len(kwargs) == 0
    id = None
    feature_sets = None
    # init file
    if isinstance(data, pd.DataFrame) or isinstance(data, ad.AnnData):
        files = None
        if isinstance(data, pd.DataFrame):
            file = File.from_df(data, run=run, description="tmp")
        elif isinstance(data, ad.AnnData):
            file = File.from_anndata(data, run=run, description="tmp")
        feature_sets = file._feature_sets  # type: ignore
        hash = file.hash  # type: ignore
        id = file.id  # type: ignore
        file.description = f"See dataset {id}"  # type: ignore
    # init files
    else:
        file = None
        if hasattr(data, "__getitem__"):
            assert isinstance(data[0], File)  # type: ignore
            files = data
            hash, feature_sets = from_files(files)  # type: ignore
        else:
            raise ValueError("Only DataFrame, AnnData and iterable of File is allowed")
    super(Dataset, dataset).__init__(
        id=id, name=name, description=description, file=file, hash=hash
    )
    dataset._files = files
    dataset._feature_sets = feature_sets


def from_files(files: Iterable[File]) -> Tuple[str, List[FeatureSet]]:
    # assert all files are already saved
    saved = not any([file._state.adding for file in files])
    if not saved:
        raise ValueError("Not all files are yet saved, please save them")
    # query all feature sets of files
    file_ids = [file.id for file in files]
    # query all feature sets at the same time rather than making a single query per file
    feature_set_file_links = File.feature_sets.through.objects.filter(
        file_id__in=file_ids
    )
    feature_set_ids = [link.feature_set_id for link in feature_set_file_links]
    feature_sets = FeatureSet.filter(id__in=feature_set_ids)
    # validate consistency of hashes
    # we do not allow duplicate hashes
    hashes = [file.hash for file in files]
    if len(hashes) != len(set(hashes)):
        seen = set()
        non_unique = [x for x in hashes if x in seen or seen.add(x)]  # type: ignore
        raise ValueError(
            "Please pass files with distinct hashes: these ones are non-unique"
            f" {non_unique}"
        )
    hash = hash_set(set(hashes))
    return hash, feature_sets


def backed(dataset: Dataset):
    if dataset.file is None:
        raise RuntimeError("Can only call backed() for datasets with a single file")
    return dataset.file.backed()


def load(dataset: Dataset):
    """Load the combined dataset."""
    if dataset.file is not None:
        return dataset.file.load()
    else:
        suffixes = [file.suffix for file in dataset.files.all()]
        if len(set(suffixes)) != 1:
            raise RuntimeError(
                "Can only load datasets where all files have the same suffix"
            )
        objects = [file.load() for file in dataset.files.all()]
        if isinstance(objects[0], pd.DataFrame):
            return pd.concat(objects)
        elif isinstance(objects[0], ad.AnnData):
            return ad.concat(objects)


def delete(dataset: Dataset, storage: bool = False):
    super(Dataset, dataset).delete()
    if dataset.file is not None:
        dataset.file.delete(storage=storage)


def save(dataset: Dataset):
    if dataset.file is not None:
        dataset.file.save()
    feature_sets = dataset._feature_sets
    if isinstance(feature_sets, dict):
        feature_sets = feature_sets.values()
    for feature_set in feature_sets:
        feature_set.save()
    super(Dataset, dataset).save()
    if dataset._files is not None and len(dataset._files) > 0:
        dataset.files.set(dataset._files)
    if len(dataset._feature_sets) > 0:
        dataset.feature_sets.set(feature_sets)


Dataset.__init__ = __init__
Dataset.from_files = from_files
Dataset.backed = backed
Dataset.load = load
Dataset.delete = delete
Dataset.save = save



"""LaminDB's core registries."""
_schema_id = "yvzi"
_name = "core"
__version__ = "0.45.5"


from lamindb_setup import _check_instance_setup

_INSTANCE_SETUP = _check_instance_setup()

if _INSTANCE_SETUP:
    from . import ids, types
    from .models import ORM  # backward compat
    from .models import (  # type: ignore
        Dataset,
        Feature,
        FeatureSet,
        File,
        Label,
        Modality,
        ParentsAware,
        Registry,
        Run,
        Storage,
        SynonymsAware,
        Transform,
        User,
        ValidationAware,
    )



from datetime import datetime
from typing import Iterable, List, NamedTuple, Optional

import pandas as pd
from django.db import models
from lamindb_setup.dev._docs import doc_args
from lnschema_core.models import Registry, SynonymsAware, ValidationAware
from lnschema_core.types import ListLike, StrField


class NoResultFound(Exception):
    pass


class MultipleResultsFound(Exception):
    pass


def format_and_convert_to_local_time(series: pd.Series):
    tzinfo = datetime.now().astimezone().tzinfo
    timedelta = tzinfo.utcoffset(datetime.now())  # type: ignore
    return (series + timedelta).dt.strftime("%Y-%m-%d %H:%M:%S")


class QuerySet(models.QuerySet):
    """Lazily loaded queried records returned by queries.

    See Also:

        `django QuerySet <https://docs.djangoproject.com/en/4.2/ref/models/querysets/>`__ # noqa

    Examples:

        >>> ln.Label(name="my label").save()
        >>> queryset = ln.Label.filter(name="my label")
        >>> queryset
        <QuerySet [Label(id=MIeZISeF, name=my label, updated_at=2023-07-19 19:53:34, created_by_id=DzTjkKse)]> # noqa
    """

    def df(self, include: Optional[List[str]] = None):
        """Convert to ``pd.DataFrame``.

        By default, shows all fields that aren't many-to-many fields, except
        ``created_at``.

        If you'd like to include many-to-many fields, use parameter ``include``.

        Args:
            include: ``Optional[List[str]] = None`` Additional (many-to-many)
                fields to include. Takes expressions like ``"labels__name"``
                ``"cell_types__name"``.

        Examples:

            >>> ln.save(ln.Label.from_values(["Label1", "Label2", "Label3"], field="name")) # noqa
            >>> ln.Label.filter().df()
                          name  external_id           updated_at  created_by_id
                  id
            wsCyIq2Z  Label1         None  2023-07-19 19:14:08       DzTjkKse
            MvpDP8Y3  Label2         None  2023-07-19 19:14:08       DzTjkKse
            zKFFabCu  Label3         None  2023-07-19 19:14:08       DzTjkKse
            >>> label = ln.Label.filter(name="Label1").one()
            >>> label = ln.Label.filter(name="benchmark").one()
            >>> label.parents.add(label)
            >>> ln.Label.filter().df(include=["labels__name", "labels__created_by_id"])
                      labels__name  labels__created_by_id      name  external_id           updated_at  created_by_id # noqa
                  id
            wsCyIq2Z  [benchmark]          [DzTjkKse]  Label1         None  2023-07-19 19:14:08       DzTjkKse # noqa
            MvpDP8Y3         None                None  Label2         None  2023-07-19 19:14:08       DzTjkKse # noqa
            zKFFabCu         None                None  Label3         None  2023-07-19 19:14:08       DzTjkKse # noqa
        """
        data = self.values()
        if len(data) > 0:
            keys = list(data[0].keys())
            if "created_at" in keys:
                keys.remove("created_at")
        else:
            keys = [
                field.name
                for field in self.model._meta.fields
                if (
                    not isinstance(field, models.ForeignKey)
                    and field.name != "created_at"
                )
            ]
            keys += [
                f"{field.name}_id"
                for field in self.model._meta.fields
                if isinstance(field, models.ForeignKey)
            ]
        df = pd.DataFrame(self.values(), columns=keys)
        if len(df) > 0 and "updated_at" in df:
            df.updated_at = format_and_convert_to_local_time(df.updated_at)
        if len(df) > 0 and "run_at" in df:
            df.run_at = format_and_convert_to_local_time(df.run_at)
        if "id" in df.columns:
            df = df.set_index("id")
        if len(df) == 0:
            return df
        if include is not None:
            if isinstance(include, str):
                include = [include]
            # fix ordering
            include = include[::-1]
            for expression in include:
                split = expression.split("__")
                field_name = split[0]
                if len(split) > 1:
                    lookup_str = "__".join(split[1:])
                else:
                    lookup_str = "id"
                Registry = self.model
                field = getattr(Registry, field_name)
                if not isinstance(field.field, models.ManyToManyField):
                    raise ValueError("Only many-to-many fields are allowed here.")
                related_ORM = (
                    field.field.model
                    if field.field.model != Registry
                    else field.field.related_model
                )
                if Registry == related_ORM:
                    left_side_link_model = f"from_{Registry.__name__.lower()}"
                    values_expression = f"to_{Registry.__name__.lower()}__{lookup_str}"
                else:
                    left_side_link_model = f"{Registry.__name__.lower()}"
                    values_expression = f"{related_ORM.__name__.lower()}__{lookup_str}"
                link_df = pd.DataFrame(
                    field.through.objects.values(
                        left_side_link_model, values_expression
                    )
                )
                link_groupby = link_df.groupby(left_side_link_model)[
                    values_expression
                ].apply(list)
                df = pd.concat((link_groupby, df), axis=1, join="inner")
                df.rename(columns={values_expression: expression}, inplace=True)
        return df

    def list(self, field: Optional[str] = None) -> List[Registry]:
        """Populate a list with the results.

        Examples:

            >>> ln.save(ln.Label.from_values(["Label1", "Label2", "Label3"], field="name")) # noqa
            >>> queryset = ln.Label.filter(name__icontains = "project")
            >>> queryset.list()
            [Label(id=NAgTZxoo, name=Label1, updated_at=2023-07-19 19:25:48, created_by_id=DzTjkKse), # noqa
            Label(id=bnsAgKRC, name=Label2, updated_at=2023-07-19 19:25:48, created_by_id=DzTjkKse), # noqa
            Label(id=R8xhAJNE, name=Label3, updated_at=2023-07-19 19:25:48, created_by_id=DzTjkKse)] # noqa
            >>> queryset.list("name")
            ['Label1', 'Label2', 'Label3']
        """
        if field is None:
            return [item for item in self]
        else:
            return [item for item in self.values_list(field, flat=True)]

    def first(self) -> Optional[Registry]:
        """If non-empty, the first result in the query set, otherwise None.

        Examples:
            >>> ln.save(ln.Label.from_values(["Label1", "Label2", "Label3"], field="name")) # noqa
            >>> queryset = ln.Label.filter(name__icontains = "project")
            >>> queryset.first()
            Label(id=NAgTZxoo, name=Label1, updated_at=2023-07-19 19:25:48, created_by_id=DzTjkKse) # noqa
        """
        if len(self) == 0:
            return None
        return self[0]

    def one(self) -> Registry:
        """Exactly one result. Throws error if there are more or none.

        Examples:
            >>> ln.Label(name="benchmark").save()
            >>> ln.Label.filter(name="benchmark").one()
            Label(id=gznl0GZk, name=benchmark, updated_at=2023-07-19 19:39:01, created_by_id=DzTjkKse) # noqa
        """
        if len(self) == 0:
            raise NoResultFound
        elif len(self) > 1:
            raise MultipleResultsFound
        else:
            return self[0]

    def one_or_none(self) -> Optional[Registry]:
        """At most one result. Returns it if there is one, otherwise returns None.

        Examples:
            >>> ln.Label(name="benchmark").save()
            >>> ln.Label.filter(name="benchmark").one_or_none()
            Label(id=gznl0GZk, name=benchmark, updated_at=2023-07-19 19:39:01, created_by_id=DzTjkKse) # noqa
            >>> ln.Label.filter(name="non existing label").one_or_none()
            None
        """
        if len(self) == 0:
            return None
        elif len(self) == 1:
            return self[0]
        else:
            raise MultipleResultsFound

    @doc_args(Registry.search.__doc__)
    def search(self, string: str, **kwargs):
        """{}"""
        from ._registry import _search

        return _search(cls=self, string=string, **kwargs)

    @doc_args(Registry.lookup.__doc__)
    def lookup(self, field: Optional[StrField] = None) -> NamedTuple:
        """{}"""
        from ._registry import _lookup

        return _lookup(cls=self, field=field)

    @doc_args(ValidationAware.validate.__doc__)
    def validate(self, values: ListLike, field: StrField, **kwargs):
        """{}"""
        from ._validate import _validate

        return _validate(cls=self, values=values, field=field, **kwargs)

    @doc_args(ValidationAware.inspect.__doc__)
    def inspect(self, values: ListLike, field: StrField, **kwargs):
        """{}"""
        from ._validate import _inspect

        return _inspect(cls=self, values=values, field=field, **kwargs)

    @doc_args(SynonymsAware.standardize.__doc__)
    def standardize(self, values: Iterable, **kwargs):
        """{}"""
        from ._synonym import _standardize

        return _standardize(cls=self, values=values, **kwargs)


setattr(models.QuerySet, "df", QuerySet.df)
setattr(models.QuerySet, "list", QuerySet.list)
setattr(models.QuerySet, "first", QuerySet.first)
setattr(models.QuerySet, "one", QuerySet.one)
setattr(models.QuerySet, "one_or_none", QuerySet.one_or_none)
setattr(models.QuerySet, "search", QuerySet.search)
setattr(models.QuerySet, "lookup", QuerySet.lookup)
setattr(models.QuerySet, "validate", QuerySet.validate)
setattr(models.QuerySet, "inspect", QuerySet.inspect)
setattr(models.QuerySet, "standardize", QuerySet.standardize)



# Generated by Django 4.2.2 on 2023-07-24 09:35

import django.db.models.deletion
from django.db import migrations, models

import lnschema_core.ids
import lnschema_core.users


class Migration(migrations.Migration):
    dependencies = [
        ("lnschema_core", "0010_dataset_categories_file_categories"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="Tag",
            new_name="Label",
        ),
        migrations.RenameField(
            model_name="featureset",
            old_name="field",
            new_name="ref_field",
        ),
        migrations.RenameField(
            model_name="featureset",
            old_name="schema",
            new_name="ref_schema",
        ),
        migrations.RenameField(
            model_name="file",
            old_name="tags",
            new_name="labels",
        ),
        migrations.RemoveField(
            model_name="dataset",
            name="categories",
        ),
        migrations.RemoveField(
            model_name="file",
            name="categories",
        ),
        migrations.AddField(
            model_name="dataset",
            name="labels",
            field=models.ManyToManyField(related_name="datasets", to="lnschema_core.label"),
        ),
        migrations.AddField(
            model_name="featureset",
            name="n",
            field=models.IntegerField(default=0),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="featureset",
            name="name",
            field=models.CharField(default=None, max_length=128, null=True),
        ),
        migrations.AddField(
            model_name="featureset",
            name="readout",
            field=models.CharField(default=None, max_length=64, null=True),
        ),
        migrations.AddField(
            model_name="featureset",
            name="ref_orm",
            field=models.CharField(db_index=True, default="Feature", max_length=64),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="label",
            name="feature",
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name="labels", to="lnschema_core.feature"),
        ),
        migrations.AddField(
            model_name="label",
            name="ref_id",
            field=models.CharField(default=None, max_length=20, null=True),
        ),
        migrations.AddField(
            model_name="label",
            name="ref_orm",
            field=models.CharField(default=None, max_length=30, null=True),
        ),
        migrations.AddField(
            model_name="label",
            name="ref_schema",
            field=models.CharField(default=None, max_length=30, null=True),
        ),
        migrations.RunSQL("UPDATE lnschema_core_feature SET type='float' where type IS NULL"),
        migrations.AlterField(
            model_name="feature",
            name="type",
            field=models.CharField(db_index=True, default=None, max_length=64),
        ),
        migrations.AlterField(
            model_name="featureset",
            name="type",
            field=models.CharField(default=None, max_length=64, null=True),
        ),
        migrations.AlterField(
            model_name="file",
            name="run",
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.PROTECT, related_name="output_files", to="lnschema_core.run"),
        ),
        migrations.AlterField(
            model_name="file",
            name="transform",
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.PROTECT, related_name="files", to="lnschema_core.transform"),
        ),
        migrations.AlterField(
            model_name="label",
            name="created_by",
            field=models.ForeignKey(
                default=lnschema_core.users.current_user_id, on_delete=django.db.models.deletion.PROTECT, related_name="created_labels", to="lnschema_core.user"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="label",
            unique_together={("name", "feature")},
        ),
        migrations.DeleteModel(
            name="Category",
        ),
    ]



import os
import shutil
import traceback
from collections import defaultdict
from datetime import datetime
from functools import partial
from typing import Iterable, List, Optional, Tuple, Union, overload  # noqa

import lamindb_setup
from django.db import transaction
from lamin_utils import logger
from lnschema_core.models import File, Registry

from lamindb.dev.storage import store_object
from lamindb.dev.storage.file import (
    auto_storage_key_from_file,
    delete_storage_using_key,
    print_hook,
)

try:
    from lamindb.dev.storage._zarr import write_adata_zarr
except ImportError:

    def write_adata_zarr(filepath):  # type: ignore
        raise ImportError("Please install zarr: pip install zarr")


def save(records: Iterable[Registry], **kwargs) -> None:  # type: ignore
    """Bulk save to registries & storage.

    Note:

        This is a much **faster** way to save many records in the database.

    Warning:

        It neither automatically creates related records nor updates existing records!
        Use ``Registry.save()`` for these use cases.

    Args:
        records: One or multiple ``Registry`` objects.

    Examples:

        Save a collection of records in one transaction, which is much faster
        than writing a loop over calls ``projects.save()``:

        >>> labels = [ln.Label(f"Label {i}") for i in range(10)]
        >>> ln.save(projects)

        For a single record, use ``.save()``:

        >>> transform = ln.Transform(name="My pipeline")
        >>> transform.save()

        Update a single existing record:

        >>> transform = ln.filter(ln.Transform, id="0Cb86EZj").one()
        >>> transform.name = "New name"
        >>> transform.save()

    """
    if isinstance(records, Iterable):
        records = set(records)
    elif isinstance(records, Registry):
        records = {records}

    # we're distinguishing between files and non-files
    # because for files, we want to bulk-upload
    # rather than upload one-by-one
    files = {r for r in records if isinstance(r, File)}
    non_files = records.difference(files)
    if non_files:
        # first save all records without recursing parents
        bulk_create(non_files)
        non_files_with_parents = {r for r in non_files if hasattr(r, "_parents")}

        if len(non_files_with_parents) > 0 and kwargs.get("parents") is not False:
            # this can only happen within lnschema_bionty right now!!
            # we might extend to core lamindb later
            import lnschema_bionty as lb

            if kwargs.get("parents") or (
                kwargs.get("parents") is None and lb.settings.auto_save_parents
            ):
                # save the record with parents one by one
                logger.warning(
                    "now recursing through parents: "
                    "this only happens once, but is much slower than bulk saving"
                )
                logger.hint(
                    "you can switch this off via: lb.settings.auto_save_parents = False"
                )
                for record in non_files_with_parents:
                    record._save_ontology_parents()

    if files:
        with transaction.atomic():
            for record in files:
                record._save_skip_storage()
        store_files(files)

    # this function returns None as potentially 10k records might be saved
    # refreshing all of them from the DB would mean a severe performance penalty
    # 2nd reason: consistency with Django Model.save(), which also returns None
    return None


def bulk_create(records: Iterable[Registry]):
    records_by_orm = defaultdict(list)
    for record in records:
        records_by_orm[record.__class__].append(record)
    for orm, records in records_by_orm.items():
        orm.objects.bulk_create(records, ignore_conflicts=True)


# This is also used within File.save()
def check_and_attempt_upload(file: File) -> Optional[Exception]:
    # if File object is either newly instantiated or replace() was called on
    # a local env it will have a _local_filepath and needs to be uploaded
    if hasattr(file, "_local_filepath"):
        try:
            upload_data_object(file)
        except Exception as exception:
            logger.warning(f"could not upload file: {file}")
            return exception
        # copies (if ob-disk) or moves the temporary file (if in-memory) to the cache
        copy_or_move_to_cache(file)
        # after successful upload, we should remove the attribute so that another call
        # call to save won't upload again, the user should call replace() then
        del file._local_filepath
    # returning None means proceed (either success or no action needed)
    return None


def copy_or_move_to_cache(file: File):
    local_path = file._local_filepath

    # in-memory zarr or not cloud
    if local_path is None or not lamindb_setup.settings.storage.is_cloud:
        return None

    local_path = local_path.resolve()

    # on-disk zarr
    if not local_path.is_file():
        return None

    # maybe create something like storage.key_to_local(key) later to simplfy
    storage_key = auto_storage_key_from_file(file)
    storage_path = lamindb_setup.settings.storage.key_to_filepath(storage_key)
    cache_path = lamindb_setup.settings.storage.cloud_to_local_no_update(storage_path)
    cache_path.parent.mkdir(parents=True, exist_ok=True)

    cache_dir = lamindb_setup.settings.storage.cache_dir

    if cache_dir in local_path.parents:
        local_path.replace(cache_path)
    else:
        shutil.copy(local_path, cache_path)
    # make sure that the cached version is older than the cloud one
    mts = datetime.now().timestamp() + 1.0
    os.utime(cache_path, times=(mts, mts))


# This is also used within File.save()
def check_and_attempt_clearing(file: File) -> Optional[Exception]:
    # this is a clean-up operation after replace() was called
    # this will only evaluate to True if replace() was called
    if hasattr(file, "_clear_storagekey"):
        try:
            if file._clear_storagekey is not None:
                delete_storage_using_key(file, file._clear_storagekey)
                logger.success(
                    f"deleted stale object at storage key {file._clear_storagekey}"
                )
                file._clear_storagekey = None
        except Exception as exception:
            return exception
    # returning None means proceed (either success or no action needed)
    return None


def store_files(files: Iterable[File]) -> None:
    """Upload files in a list of database-committed files to storage.

    If any upload fails, subsequent files are cleaned up from the DB.
    """
    exception: Optional[Exception] = None
    # because uploads might fail, we need to maintain a new list
    # of the succeeded uploads
    stored_files = []

    # upload new local files
    for file in files:
        exception = check_and_attempt_upload(file)
        if exception is not None:
            break
        stored_files += [file]
        exception = check_and_attempt_clearing(file)
        if exception is not None:
            logger.warning(f"clean up of {file._clear_storagekey} failed")
            break

    if exception is not None:
        # clean up metadata for files not uploaded to storage
        with transaction.atomic():
            for file in files:
                if file not in stored_files:
                    file._delete_skip_storage()
        error_message = prepare_error_message(files, stored_files, exception)
        raise RuntimeError(error_message)
    return None


def prepare_error_message(records, stored_files, exception) -> str:
    if len(records) == 1 or len(stored_files) == 0:
        error_message = (
            "No entries were uploaded or committed"
            " to the database. See error message:\n\n"
        )
    else:
        error_message = (
            "The following entries have been"
            " successfully uploaded and committed to the database:\n"
        )
        for record in stored_files:
            error_message += (
                f"- {', '.join(record.__repr__().split(', ')[:3]) + ', ...)'}\n"
            )
        error_message += "\nSee error message:\n\n"
    error_message += f"{str(exception)}\n\n{traceback.format_exc()}"
    return error_message


def upload_data_object(file) -> None:
    """Store and add file and its linked entries."""
    # do NOT hand-craft the storage key!
    file_storage_key = auto_storage_key_from_file(file)
    msg = f"storing file '{file.id}' at '{file_storage_key}'"
    if hasattr(file, "_to_store") and file._to_store and file.suffix != ".zarr":
        logger.save(msg)
        store_object(file._local_filepath, file_storage_key)
    elif (
        file.suffix in {".zarr", ".zrad"}
        and hasattr(file, "_memory_rep")
        and file._memory_rep is not None
    ):
        logger.save(msg)
        storagepath = lamindb_setup.settings.storage.key_to_filepath(file_storage_key)
        print_progress = partial(
            print_hook, filepath=file_storage_key, action="uploading"
        )
        write_adata_zarr(file._memory_rep, storagepath, callback=print_progress)



# Generated by Django 4.2.2 on 2023-08-10 11:36

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lnschema_core", "0014_rename_ref_field_featureset_registry"),
    ]

    operations = [
        migrations.AddField(
            model_name="file",
            name="initial_version",
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.PROTECT, to="lnschema_core.file"),
        ),
        migrations.AddField(
            model_name="file",
            name="version",
            field=models.CharField(db_index=True, default=None, max_length=10, null=True),
        ),
    ]



import pytest

import lamindb_setup as ln_setup


@pytest.fixture
def setup_instance():
    ln_setup.init(storage="./testdb")
    yield
    ln_setup.delete("testdb", force=True)


# tested in notebook right now
# def test_migrate_create(setup_instance):
#     assert ln_setup.migrate.create() is None
# def test_migrate_deploy(setup_instance):
#     assert ln_setup.migrate.deploy() is None


def test_migrate_check(setup_instance):
    assert ln_setup.migrate.check()



import builtins
import os
import re
from datetime import datetime, timezone
from pathlib import Path, PurePath
from typing import Dict, List, Optional, Tuple, Union

import lnschema_core
from lamin_utils import logger
from lamindb_setup import settings
from lamindb_setup.dev import InstanceSettings
from lnschema_core import Run, Transform
from lnschema_core.types import TransformType

is_run_from_ipython = getattr(builtins, "__IPYTHON__", False)

msg_path_failed = (
    "failed to infer notebook path.\nfix: either track manually via"
    " `ln.track(ln.Transform(name='My notebook'))` or pass"
    " `notebook_path` to ln.track()"
)

msg_manual_init = (
    "\n(1) save your notebook!"
    "\n(2) attach metadata to the notebook by running the CLI:\n"
    "lamin track {notebook_path}"
    "\n(3) reload or re-open your notebook"
)


class UpdateNbWithNonInteractiveEditorError(Exception):
    pass


class NotebookNotSavedError(Exception):
    pass


class NoTitleError(Exception):
    pass


def _seconds_modified(filepath):
    return datetime.now().timestamp() - Path(filepath).stat().st_mtime


def _write_notebook_meta(metadata):
    from nbproject import dev as nb_dev
    from nbproject._header import _env, _filepath

    nb_dev._frontend_commands._save_notebook(_env)
    nb = nb_dev.read_notebook(_filepath)
    nb.metadata["nbproject"] = metadata

    # write proper execution count
    header_re = re.compile(r"^[^#]*track\(", flags=re.MULTILINE)
    ccount = 0
    for cell in nb.cells:
        if cell["cell_type"] != "code":
            continue
        elif cell["execution_count"] is not None:
            ccount = cell["execution_count"]
        if header_re.match("".join(cell["source"])) is not None:
            cell["execution_count"] = ccount + 1
            break

    nb_dev.write_notebook(nb, _filepath)
    nb_dev._frontend_commands._reload_notebook(_env)


def reinitialize_notebook(
    transform: Transform, metadata: Optional[Dict] = None
) -> Tuple[Transform, Dict]:
    from nbproject import dev as nb_dev
    from nbproject._header import _filepath

    new_id, new_version = transform.stem_id, None

    if "NBPRJ_TEST_NBPATH" not in os.environ:
        response = input("Do you want to generate a new id? (y/n)")
    else:
        response = "y"
    if response == "y":
        new_id = lnschema_core.ids.base62_12()
    else:
        response = input("Do you want to set a new version (e.g. '1.1')? (y/n)")
        if response == "y":
            new_version = input("Please type the version: ")

    nb = None
    if metadata is None:
        nb = nb_dev.read_notebook(_filepath)
        metadata = nb.metadata["nbproject"]

    metadata["id"] = new_id
    if new_version is None:
        new_version = "0"
    metadata["version"] = new_version

    # here we check that responses to both inputs (for new id and version) were not 'n'
    if transform.stem_id != new_id or transform.version != new_version:
        transform = Transform(
            stem_id=new_id, version=new_version, type=TransformType.notebook
        )
    return transform, metadata


# from https://stackoverflow.com/questions/61901628
def get_notebook_name_colab() -> str:
    from socket import gethostbyname, gethostname  # type: ignore

    from requests import get  # type: ignore

    ip = gethostbyname(gethostname())  # 172.28.0.12
    name = get(f"http://{ip}:9000/api/sessions").json()[0]["name"]
    return name.rstrip(".ipynb")


class run_context:
    """Global run context."""

    instance: Optional[InstanceSettings] = None
    """Current instance."""
    transform: Optional[Transform] = None
    """Current transform."""
    run: Optional[Run] = None
    """Current run."""

    # exposed to user as ln.track()
    @classmethod
    def _track(
        cls,
        transform: Optional[Transform] = None,
        *,
        new_run: Optional[bool] = None,
        reference: Optional[str] = None,
        reference_type: Optional[str] = None,
        notebook_path: Optional[str] = None,
        pypackage: Optional[Union[str, List[str]]] = None,
        editor: Optional[str] = None,
    ) -> None:
        """Track global `Transform` & `Run` for a notebook or pipeline.

        Creates or loads a :class:`~lamindb.Run` record and sets a global
        :class:`~lamindb.dev.run_context`.

        In a Jupyter notebook, call without any argument (metadata is parsed).
        If the notebook has no associated metadata ("is not initialized"),
        attempts to write metadata to disk. If it fails to so interactively, it
        will ask you to leverage the CLI.

        Args:
            transform: Can be of type `"pipeline"` or `"notebook"`
                (:class:`~lamindb.types.TransformType`).
            new_run: If `False`, loads latest run of transform
                (default notebook), if True, creates new run (default pipeline).
            reference: Reference to pass to :class:`~lamindb.Run` record.
            reference_type: Reference type to pass to :class:`~lamindb.Run` record (e.g. "url").
            notebook_path: Filepath of notebook. Only needed if inference fails.
            pypackage: One or more python packages for which to parse versions.
            editor: Editor environment.
                Pass `'lab'` for jupyter lab and `'notebook'` for jupyter notebook,
                this can help to identify the correct mechanism for interactivity
                when automatic inference fails.

        Examples:

            If you're in a Jupyter notebook and installed lamindb with `pip
            install[jupyter]`, you can simply call:

            >>> ln.track()
            ✅ saved: Transform(id=1LCd8kco9lZUBg, name=Track data lineage / provenance, short_name=02-data-lineage, stem_id=1LCd8kco9lZU, version=0, type=notebook, updated_at=2023-07-10 18:37:19, created_by_id=DzTjkKse) # noqa
            ✅ saved: Run(id=pHgVICV9DxBaV6BAuKJl, run_at=2023-07-10 18:37:19, transform_id=1LCd8kco9lZUBg, created_by_id=DzTjkKse) # noqa
            >>> ln.dev.run_context.transform
            Transform(id=1LCd8kco9lZUBg, name=Track data lineage / provenance, short_name=02-data-lineage, stem_id=1LCd8kco9lZU, version=0, type=notebook, updated_at=2023-07-10 18:37:19, created_by_id=DzTjkKse) # noqa
            >>> ln.dev.run_context.run
            Run(id=pHgVICV9DxBaV6BAuKJl, run_at=2023-07-10 18:37:19, transform_id=1LCd8kco9lZUBg, created_by_id=DzTjkKse) # noqa

            If you'd like to track a pipeline we need to pass a
            :class:`~lamindb.Transform` object of `type` `"pipeline"`:

            >>> ln.Transform(name="Cell Ranger", version="7.2.0", type="pipeline").save()
            >>> transform = ln.Transform.filter(name="Cell Ranger", version="7.2.0").one()
            >>> ln.track(transform)
            💬 loaded: Transform(id=ceHkZMaiHFdoB6, name=Cell Ranger, stem_id=ceHkZMaiHFdo, version=7.2.0, type=pipeline, updated_at=2023-07-10 18:37:19, created_by_id=DzTjkKse) # noqa
            ✅ saved: Run(id=RcpWIKC8cF74Pn3RUJ1W, run_at=2023-07-10 18:37:19, transform_id=ceHkZMaiHFdoB6, created_by_id=DzTjkKse) # noqa
            >>> ln.context.transform
            Transform(id=ceHkZMaiHFdoB6, name=Cell Ranger, stem_id=ceHkZMaiHFdo, version=7.2.0, type=pipeline, updated_at=2023-07-10 18:37:19, created_by_id=DzTjkKse) # noqa
            >>> ln.context.run
            Run(id=RcpWIKC8cF74Pn3RUJ1W, run_at=2023-07-10 18:37:19, transform_id=ceHkZMaiHFdoB6, created_by_id=DzTjkKse) # noqa
        """
        cls.instance = settings.instance
        import lamindb as ln

        if transform is None:
            is_tracked_notebook = False

            if is_run_from_ipython:
                try:
                    cls._track_notebook(
                        pypackage=pypackage,
                        filepath=notebook_path,
                        editor=editor,
                    )
                    is_tracked_notebook = True
                except Exception as e:
                    if isinstance(e, ImportError):
                        logger.info(
                            "it looks like you are running ln.track() from a "
                            "notebook!\nplease install nbproject: pip install nbproject"
                        )
                    elif isinstance(e, UpdateNbWithNonInteractiveEditorError):
                        raise e
                    elif isinstance(e, (NotebookNotSavedError, NoTitleError)):
                        raise e
                    else:
                        logger.warning(f"automatic tracking of notebook failed: {e}")
                    is_tracked_notebook = False

            if not is_tracked_notebook:
                logger.warning(
                    "no automatic metadata detection, consider passing transform"
                )
                return None
        else:
            transform_exists = None
            if transform.id is not None:
                # transform has an id but unclear whether already saved
                transform_exists = Transform.filter(id=transform.id).first()
            if transform_exists is None:
                transform.save()
                logger.save(f"saved: {transform}")
                transform_exists = transform
            else:
                logger.success(f"loaded: {transform_exists}")
            cls.transform = transform_exists

        if new_run is None:  # for notebooks, default to loading latest runs
            new_run = False if cls.transform.type == TransformType.notebook.value else True  # type: ignore  # noqa

        run = None
        if not new_run:  # try loading latest run by same user
            run = (
                ln.Run.filter(
                    transform=cls.transform, created_by_id=ln.setup.settings.user.id
                )
                .order_by("-created_at")
                .first()
            )
            if run is not None:  # loaded latest run
                run.run_at = datetime.now(timezone.utc)  # update run time
                run.reference = reference
                run.reference_type = reference_type
                run.save()
                logger.success(f"loaded: {run}")

        if run is None:  # create new run
            run = ln.Run(
                transform=cls.transform,
                reference=reference,
                reference_type=reference_type,
            )
            run.save()
            logger.save(f"saved: {run}")
        cls.run = run

        # at this point, we have a transform can display its parents if there are any
        parents = cls.transform.parents.all() if cls.transform is not None else []
        if len(parents) > 0:
            if len(parents) == 1:
                logger.info(f"  parent transform: {parents[0]}")
            else:
                parents_formatted = "\n   - ".join([f"{parent}" for parent in parents])
                logger.info(f"  parent transforms:\n   - {parents_formatted}")

        # only for newly intialized notebooks
        if hasattr(cls, "_notebook_meta"):
            _write_notebook_meta(cls._notebook_meta)  # type: ignore
            del cls._notebook_meta  # type: ignore

    @classmethod
    def _track_notebook(
        cls,
        *,
        filepath: Optional[str] = None,
        pypackage: Optional[Union[str, List[str]]] = None,
        editor: Optional[str] = None,
    ):
        """Infer Jupyter notebook metadata and create `Transform` record.

        Args:
            pypackage: One or more python packages to track.
            filepath: Filepath of notebook. Only needed if automatic inference fails.
            editor: Editor environment. Only needed if automatic inference fails.
                Pass `'lab'` for jupyter lab and `'notebook'` for jupyter notebook,
                this can help to identify the correct mechanism for interactivity
                when automatic inference fails.
        """
        import nbproject
        from nbproject.dev._jupyter_communicate import notebook_path

        cls.instance = settings.instance

        metadata = None
        needs_init = False
        is_interactive = False
        reference = None
        if filepath is None:
            path_env = None
            try:
                path_env = notebook_path(return_env=True)
            except Exception:
                raise RuntimeError(msg_path_failed)
            if path_env is None:
                raise RuntimeError(msg_path_failed)
            notebook_path, _env = path_env
        else:
            notebook_path, _env = filepath, editor
        if isinstance(notebook_path, (Path, PurePath)):
            notebook_path = notebook_path.as_posix()
        if notebook_path.endswith("Untitled.ipynb"):
            raise RuntimeError("Please rename your notebook before tracking it")
        if notebook_path.startswith("/fileId="):
            # This is Google Colab!
            # google colab fileID looks like this
            # /fileId=1KskciVXleoTeS_OGoJasXZJreDU9La_l
            # we'll take the first 12 characters
            colab_id = notebook_path.replace("/fileId=", "")
            id = colab_id[:12]
            reference = f"colab_id: {colab_id}"
            filestem = get_notebook_name_colab()
            _env = "colab"
        else:
            try:
                metadata, needs_init, nb = nbproject.header(
                    pypackage=pypackage,
                    filepath=notebook_path if filepath is None else filepath,
                    env=_env if editor is None else editor,
                    metadata_only=True,
                )
                # this contains filepath if the header was run successfully
                from nbproject._header import _env, _filepath  # type: ignore
            except Exception as e:
                nbproject_failed_msg = (
                    "Auto-retrieval of notebook name & title failed.\n\nFixes: Either"
                    f" init on the CLI `lamin track {notebook_path}` or pass"
                    " transform manually `ln.track(ln.Transform(name='My"
                    " notebook'))`\n\nPlease consider pasting error at:"
                    f" https://github.com/laminlabs/nbproject/issues/new\n\n{e}"
                )
                raise RuntimeError(nbproject_failed_msg)
            try:
                from nbproject.dev._metadata_display import DisplayMeta
                from nbproject.dev._pypackage import infer_pypackages

                dm = DisplayMeta(metadata)
                logger.info(
                    "notebook imports:"
                    f" {' '.join(dm.pypackage(infer_pypackages(nb, pin_versions=True)))}"  # noqa
                )
            except Exception:
                logger.debug("inferring imported packages failed")
                pass

        if needs_init:
            if _env in ("lab", "notebook"):
                cls._notebook_meta = metadata  # type: ignore
            else:
                msg = msg_manual_init.format(notebook_path=notebook_path)
                raise UpdateNbWithNonInteractiveEditorError(msg)

        if _env in ("lab", "notebook"):
            # save the notebook in case that title was updated
            # but notebook not saved
            nbproject.dev._frontend_commands._save_notebook(_env)
            # check here if interactivity really works

        if metadata is not None:
            # this only executed if nbproject is used
            # check here if interactivity really works
            # timestamp should be recent due to save
            is_interactive = _seconds_modified(_filepath) < 1.5  # should be ~1 sec
            if not is_interactive and needs_init:
                msg = msg_manual_init.format(notebook_path=_filepath)
                raise UpdateNbWithNonInteractiveEditorError(msg)

            id = metadata["id"]
            version = metadata["version"]
            filestem = Path(_filepath).stem
            try:
                title = nbproject.meta.live.title
            except IndexError:
                raise NotebookNotSavedError(
                    "The notebook is not saved, please save the notebook and"
                    " rerun `ln.track()`"
                )
            if title is None:
                raise NoTitleError(
                    "Please add a title to your notebook in a markdown cell: # Title"
                )
        else:
            version = "0"
            title = filestem

        transform = Transform.filter(stem_id=id, version=version).one_or_none()
        if transform is None:
            transform = Transform(
                stem_id=id,
                version=version,
                name=title,
                short_name=filestem,
                reference=reference,
                type=TransformType.notebook,
            )
            transform.save()
            logger.save(f"saved: {transform}")
        else:
            logger.success(f"loaded: {transform}")
            if transform.name != title or transform.short_name != filestem:
                response = input(
                    "Updated notebook name and/or title: Do you want to assign a"
                    " new id or version? (y/n)"
                )
                if response == "y":
                    if _env in ("lab", "notebook") and is_interactive:
                        transform, metadata = reinitialize_notebook(transform, metadata)
                        # only write metadata back to notebook if it actually changed!
                        # if filename or title changed, this does not merit a write!
                        # it's dangerous to write unnecessarily
                        cls._notebook_meta = metadata  # type: ignore
                    else:
                        msg = msg_manual_init.format(notebook_path=notebook_path)
                        raise UpdateNbWithNonInteractiveEditorError(msg)
                transform.name = title
                transform.short_name = filestem
                transform.save()
                if response == "y":
                    logger.save(f"saved: {transform}")
                else:
                    logger.success(f"updated: {transform}")

        cls.transform = transform



import lnschema_bionty as lb
import pytest

import lamindb as ln
from lamindb._query_set import MultipleResultsFound, NoResultFound


def test_df():
    # for self-referential models
    project_label = ln.Label(name="Project")
    project_label.save()
    project_names = [f"Project {i}" for i in range(3)]
    labels = [ln.Label(name=name) for name in project_names]
    ln.save(labels)
    for label in labels:
        label.parents.add(project_label)
    df = ln.Label.filter().df(include="parents__name")
    assert df.columns[0] == "parents__name"
    # order is not conserved
    assert df["parents__name"][0] == [project_label.name]
    # pass a list
    df = ln.Label.filter().df(include=["parents__name", "parents__created_by_id"])
    assert df.columns[1] == "parents__created_by_id"
    assert df["parents__name"][0] == [project_label.name]
    assert set(df["parents__created_by_id"][0]) == set([ln.setup.settings.user.id])

    # for other models
    feature_names = [f"Feature {i}" for i in range(3)]
    features = [ln.Feature(name=name, type=int) for name in feature_names]
    feature_set = ln.FeatureSet(features, name="my feature_set")
    feature_set.save()
    feature_set.features.set(features)

    df = ln.FeatureSet.filter(name="my feature_set").df(include="features__name")
    assert df.columns[0] == "features__name"
    # order is not conserved
    assert set(df["features__name"][0]) == set(feature_names)
    # pass a list
    df = ln.FeatureSet.filter(name="my feature_set").df(
        include=["features__name", "features__created_by_id"]
    )
    assert df.columns[1] == "features__created_by_id"
    assert set(df["features__name"][0]) == set(feature_names)
    assert set(df["features__created_by_id"][0]) == set([ln.setup.settings.user.id])

    # raise error for non many-to-many
    with pytest.raises(ValueError):
        ln.Label.filter().df(include="name")

    # clean up
    project_label.delete()
    for label in labels:
        label.delete()

    # call it from a non-select-derived queryset
    qs = ln.User.objects.all()
    assert qs.df().iloc[0]["handle"] == "testuser1"


def test_one_first():
    qs = ln.User.objects.all()
    assert qs.one().handle == "testuser1"
    assert qs.first().handle == "testuser1"
    assert qs.one_or_none().handle == "testuser1"

    qs = ln.User.filter(handle="test")
    with pytest.raises(NoResultFound):
        qs.one()
    qs = lb.BiontySource.filter().all()
    with pytest.raises(MultipleResultsFound):
        qs.one()
    with pytest.raises(MultipleResultsFound):
        qs.one_or_none()


def test_search():
    label_names = [f"Label {i}" for i in range(3)]
    labels = [ln.Label(name=name) for name in label_names]
    ln.save(labels)
    qs = ln.Label.filter(name="Label 2").all()
    assert qs.search("Label 1").iloc[0].name == "Label 2"
    assert qs.search("Label 1", field=ln.Label.name).iloc[0].name == "Label 2"
    for label in labels:
        label.delete()


def test_lookup():
    qs = ln.User.filter(handle="testuser1").all()
    lookup = qs.lookup(field="handle")
    assert lookup.testuser1.handle == "testuser1"
    lookup = qs.lookup(field=ln.User.handle)
    assert lookup.testuser1.handle == "testuser1"


def test_inspect():
    qs = ln.User.filter(handle="testuser1").all()
    assert qs.inspect(["user1", "user2"], "name")["validated"] == []
    assert ln.User.inspect(["user1", "user2"], "name")["validated"] == []
    assert ln.User.inspect(["user1", "user2"], ln.User.name)["validated"] == []
    assert ln.User.inspect("user1", "name")["validated"] == []


def test_validate():
    qs = ln.User.filter(handle="testuser1").all()
    assert qs.validate(["testuser1", "Test User1"], "handle").tolist() == [True, False]
    assert ln.User.validate(["testuser1", "Test User1"], "handle").tolist() == [
        True,
        False,
    ]
    assert ln.User.validate(["testuser1", "Test User1"], ln.User.handle).tolist() == [
        True,
        False,
    ]
    assert ln.User.validate("testuser1", ln.User.handle).tolist() == [True]


def test_map_synonyms():
    qs = ln.User.filter(handle="testuser1").all()
    assert qs.standardize(["user1", "user2"]) == ["user1", "user2"]



from datetime import datetime, timezone
from pathlib import Path
from typing import Optional, Union

import fsspec
from dateutil.parser import isoparse  # type: ignore
from lamin_utils import logger

from ._settings_instance import InstanceSettings
from .upath import UPath, infer_filesystem

EXPIRATION_TIME = 7 * 24 * 60 * 60  # 1 week

MAX_MSG_COUNTER = 100  # print the msg after this number of iterations


class empty_locker:
    has_lock = True

    @classmethod
    def lock(cls):
        pass

    @classmethod
    def unlock(cls):
        pass


class Locker:
    def __init__(self, user_id: str, storage_root: Union[UPath, Path], name: str):
        logger.debug(f"init cloud sqlite locker: {user_id}, {storage_root}, {name}.")

        self._counter = 0

        self.user = user_id
        self.name = name

        self.root = storage_root
        self.fs, _ = infer_filesystem(storage_root)

        exclusion_path = storage_root / f"exclusion/{name}"
        self.mapper = fsspec.FSMap(str(exclusion_path), self.fs, create=True)

        priorities_path = str(exclusion_path / "priorities")
        if self.fs.exists(priorities_path):
            self.users = self.mapper["priorities"].decode().split("*")

            if self.user not in self.users:
                self.priority = len(self.users)
                self.users.append(self.user)
                # potential problem here if 2 users join at the same time
                # can be avoided by using separate files for each user
                # and giving priority by timestamp
                # here writing the whole list back because gcs
                # does not support the append mode
                self.mapper["priorities"] = "*".join(self.users).encode()
            else:
                self.priority = self.users.index(self.user)
        else:
            self.mapper["priorities"] = self.user.encode()
            self.users = [self.user]
            self.priority = 0

        self.mapper[f"numbers/{self.user}"] = b"0"
        self.mapper[f"entering/{self.user}"] = b"0"

        # clean up failures
        for user in self.users:
            for endpoint in ("numbers", "entering"):
                user_endpoint = f"{endpoint}/{user}"
                user_path = str(exclusion_path / user_endpoint)
                if not self.fs.exists(user_path):
                    continue
                if self.mapper[user_endpoint] == b"0":
                    continue
                period = (datetime.now() - self.modified(user_path)).total_seconds()
                if period > EXPIRATION_TIME:
                    logger.info(
                        f"the lock of the user {user} seems to be stale, clearing"
                        f" {endpoint}."
                    )
                    self.mapper[user_endpoint] = b"0"

        self._has_lock = None
        self._locked_by = None

    def modified(self, path):
        if "gcs" not in self.fs.protocol:
            mtime = self.fs.modified(path)
        else:
            stat = self.fs.stat(path)
            if "updated" in stat:
                mtime = stat["updated"]
                mtime = isoparse(mtime)
            else:
                return None
        # always convert to the local timezone before returning
        # assume in utc if the time zone is not specified
        if mtime.tzinfo is None:
            mtime = mtime.replace(tzinfo=timezone.utc)
        return mtime.astimezone().replace(tzinfo=None)

    def _msg_on_counter(self, user):
        if self._counter == MAX_MSG_COUNTER:
            logger.warning(f"competing for the lock with the user {user}.")

        if self._counter <= MAX_MSG_COUNTER:
            self._counter += 1

    def _lock_unsafe(self):
        if self._has_lock:
            return None

        self._has_lock = True
        self._locked_by = self.user

        self.users = self.mapper["priorities"].decode().split("*")

        self.mapper[f"entering/{self.user}"] = b"1"

        numbers = [int(self.mapper[f"numbers/{user}"]) for user in self.users]
        number = 1 + max(numbers)
        self.mapper[f"numbers/{self.user}"] = str(number).encode()

        self.mapper[f"entering/{self.user}"] = b"0"

        for i, user in enumerate(self.users):
            if i == self.priority:
                continue

            while self.mapper[f"entering/{user}"] == b"1":
                self._msg_on_counter(user)

            c_number = int(self.mapper[f"numbers/{user}"])

            if c_number == 0:
                continue

            if (number > c_number) or (number == c_number and self.priority > i):
                self._has_lock = False
                self._locked_by = user
                self.mapper[f"numbers/{self.user}"] = b"0"
                logger.warning(f"the instance is already locked by the user {user}.")
                return None

    def lock(self):
        try:
            self._lock_unsafe()
        except BaseException as e:
            self.unlock()
            self._clear()
            raise e

    def unlock(self):
        self.mapper[f"numbers/{self.user}"] = b"0"
        self._has_lock = None
        self._locked_by = None
        self._counter = 0

    def _clear(self):
        self.mapper[f"entering/{self.user}"] = b"0"

    @property
    def has_lock(self):
        if self._has_lock is None:
            logger.info("the lock has not been initialized, trying to obtain the lock.")
            self.lock()

        return self._has_lock


_locker: Optional[Locker] = None


def get_locker(isettings: InstanceSettings) -> Locker:
    from .._settings import settings

    global _locker

    user_id = settings.user.id
    storage_root = isettings.storage.root
    instance_name = isettings.name

    if (
        _locker is None
        or _locker.user != user_id
        or _locker.root is not storage_root
        or _locker.name != instance_name
    ):
        _locker = Locker(user_id, storage_root, instance_name)

    return _locker



import lamindb as ln


def test_delete():
    labels = ln.Label.from_values(["label1", "label2", "label3"], field="name")
    ln.save(labels)
    ln.delete(labels[0])
    ln.delete([labels[1], labels[2]])
    assert ln.Label.filter(name__in=["label1", "label2", "label3"]).count() == 0



import pytest

import lamindb as ln


def test_create():
    transform = ln.Transform(stem_id="NJvdsWWbJlZS", version="0")
    assert transform.id == "NJvdsWWbJlZSz8"
    transform = ln.Transform(id="wNWf6CCdoafkz8")
    assert transform.stem_id == "wNWf6CCdoafk"
    with pytest.raises(ValueError):
        ln.Transform(version=1)



import anndata as ad
import numpy as np

import lamindb as ln

adata = ad.AnnData(
    X=np.array([[1, 2, 3], [4, 5, 6]]),
    obs=dict(Obs=["A", "B"]),
    var=dict(Feat=["a", "b", "c"]),
    obsm=dict(X_pca=np.array([[1, 2], [3, 4]])),
)


def test_create_delete_from_single_dataframe():
    df = ln.dev.datasets.df_iris_in_meter_batch1()

    # try to create and save the dataset
    dataset = ln.Dataset(df, name="Iris flower dataset1")
    # because features weren't registered, there is no linked feature set
    assert dataset._feature_sets == {}
    # register features and then repeat
    ln.save(ln.Feature.from_df(df))
    dataset = ln.Dataset(df, name="Iris flower dataset1")
    assert "columns" in dataset._feature_sets

    dataset.save()

    # basics
    assert dataset.load().iloc[0].tolist() == df.iloc[0].tolist()
    file = dataset.file
    assert file.description == f"See dataset {dataset.id}"
    assert dataset.hash == file.hash
    assert dataset.id == file.id
    assert ln.File.filter(id=dataset.id).one_or_none() is not None
    assert ln.File.filter(id=file.id).one_or_none() is not None

    # features
    feature_list = [
        "sepal_length",
        "sepal_width",
        "petal_length",
        "petal_width",
        "iris_species_name",
    ]
    assert len(ln.Feature.filter(name__in=feature_list).list()) == 5
    feature_set = ln.FeatureSet.filter(datasets=dataset).one()
    feature_list_queried = ln.Feature.filter(feature_sets=feature_set).list()
    feature_list_queried = [feature.name for feature in feature_list_queried]
    assert set(feature_list_queried) == set(feature_list)
    # the feature_set is also linked to the file
    assert ln.FeatureSet.filter(files=dataset.file).one() == feature_set

    # now proceed to deletion
    dataset.delete(storage=True)
    assert ln.File.filter(id=dataset.id).one_or_none() is None
    assert ln.File.filter(id=file.id).one_or_none() is None


# def test_create_delete_from_single_anndata():
#     dataset = ln.Dataset(adata, name="My adata")
#     dataset.save()
#     assert dataset.load().iloc[0].tolist() == df.iloc[0].tolist()
#     file = dataset.file
#     assert file.description is None
#     assert dataset.hash == file.hash
#     assert dataset.id == file.id
#     assert ln.File.filter(id=dataset.id).one_or_none() is not None
#     assert ln.File.filter(id=file.id).one_or_none() is not None
#     dataset.delete(storage=True)
#     assert ln.File.filter(id=dataset.id).one_or_none() is None
#     assert ln.File.filter(id=file.id).one_or_none() is None



import secrets
import string
from pathlib import Path
from typing import Mapping, Optional, Union
from uuid import UUID

from pydantic import BaseModel, validator
from pydantic.networks import MultiHostDsn

from .upath import UPath


def base62(n_char: int) -> str:
    """Like nanoid without hyphen and underscore."""
    alphabet = string.digits + string.ascii_letters.swapcase()
    id = "".join(secrets.choice(alphabet) for i in range(n_char))
    return id


def secret() -> str:
    """Password or secret: 40 base62."""
    return base62(n_char=40)


def validate_schema_arg(schema: Optional[str] = None) -> str:
    if schema is None or schema == "":
        return ""
    # currently no actual validation, can add back if we see a need
    # the following just strips white spaces
    to_be_validated = [s.strip() for s in schema.split(",")]
    return ",".join(to_be_validated)


def validate_db_arg(db: Optional[str]) -> None:
    if db:
        LaminDsnModel(db=db)


def validate_unique_sqlite(
    *, hub, db: Optional[str], storage_id: UUID, name: str, account: Mapping
) -> None:
    # if a remote sqlite instance, make sure there is no other instance
    # that has the same name and storage location
    if db is None:  # remote sqlite instance
        instances = (
            hub.table("instance")
            .select("*")
            .eq("storage_id", storage_id)
            .eq("name", name)
            .execute()
            .data
        )
        if len(instances) > 0:
            # retrieve account owning the first instance
            accounts = (
                hub.table("account")
                .select("*")
                .eq("id", instances[0]["account_id"])
                .execute()
                .data
            )
            raise RuntimeError(
                "\nThere is already an sqlite instance with the same name and storage"
                f" location from account {accounts[0]['handle']}\nTwo sqlite instances"
                " with the same name and the same storage cannot exist\nFix: "
                f"Choose another name or load instance {accounts[0]['handle']}/{name}\n"
            )


def get_storage_region(storage_root: Union[str, Path, UPath]) -> Optional[str]:
    storage_root_str = str(storage_root)
    storage_region = None

    if storage_root_str.startswith("s3://"):
        import botocore.session as session
        from botocore.config import Config
        from botocore.exceptions import ClientError, NoCredentialsError

        bucket = storage_root_str.replace("s3://", "")

        s3_session = session.get_session()

        s3_client = s3_session.create_client("s3")
        try:
            response = s3_client.head_bucket(Bucket=bucket)
            exc = None
        except Exception as e:
            response = None
            exc = e

        if isinstance(exc, NoCredentialsError):
            s3_client = s3_session.create_client(
                "s3", config=Config(signature_version=session.UNSIGNED)
            )
            try:
                response = s3_client.head_bucket(Bucket=bucket)
                exc = None
            except Exception as e:
                response = None
                exc = e

        if isinstance(exc, ClientError):
            response = exc.response

        if response is not None:
            storage_region = (
                response["ResponseMetadata"]
                .get("HTTPHeaders", {})
                .get("x-amz-bucket-region")
            )
            return storage_region

        if exc is not None:
            raise exc

    return storage_region


def validate_storage_root_arg(storage_root: str) -> None:
    if storage_root.endswith("/"):
        raise ValueError("Pass settings.storage.root_as_str rather than path")
    if storage_root.startswith(("gs://", "s3://")):
        return None
    else:  # local path
        try:
            _ = Path(storage_root)
            return None
        except Exception:
            raise ValueError(
                "`storage` is neither a valid local, a Google Cloud nor an S3 path."
            )


def get_storage_type(storage_root: str):
    if str(storage_root).startswith("s3://"):
        return "s3"
    elif str(storage_root).startswith("gs://"):
        return "gs"
    else:
        return "local"


class LaminDsn(MultiHostDsn):
    """Custom DSN Type for Lamin.

    This class allows us to customize the allowed schemes for databases
    and also handles the parsing and building of DSN strings with the
    database name instead of URL path.
    """

    allowed_schemes = {
        "postgresql",
        # future enabled schemes
        # "snowflake",
        # "bigquery"
    }
    user_required = True
    __slots__ = ()

    @property
    def database(self):
        return self.path[1:]

    @classmethod
    def build(
        cls,
        *,
        scheme: str,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: str,
        port: Optional[str] = None,
        database: Optional[str] = None,
        query: Optional[str] = None,
        fragment: Optional[str] = None,
        **_kwargs: str,
    ) -> str:
        return super().build(
            scheme=scheme,
            user=user,
            password=password,
            host=host,
            port=port,
            path=f"/{database}",
            query=query,
            fragment=fragment,
        )


class LaminDsnModel(BaseModel):
    db: LaminDsn

    @validator("db")
    def check_db_name(cls, v):
        assert v.path and len(v.path) > 1, "database must be provided"
        return v



from lamin_utils import logger

from ._settings import settings


def register():
    """Register an instance on the hub."""
    from .dev._hub_core import init_instance as init_instance_hub

    isettings = settings.instance
    result = init_instance_hub(
        owner=isettings.owner,
        name=isettings.name,
        storage=isettings.storage.root_as_str,
        db=isettings.db if isettings.dialect != "sqlite" else None,
        schema=isettings._schema_str,
    )
    if result == "instance-exists-already":
        logger.info("instance was already registered")
    elif isinstance(result, str):
        raise RuntimeError(f"creating instance on hub failed:\n{result}")
    else:
        logger.save(f"instance registered: https://lamin.ai/{isettings.identifier}")
